# coding=utf-8
'''
lightgbm二分类问题的建模, 贝叶斯优化.
'''

import numpy as np
import pandas as pd
from scipy.stats import spearmanr
import matplotlib.pyplot as plt

from sklearn.metrics import roc_auc_score
from sklearn.model_selection import train_test_split
from sklearn.utils.class_weight import compute_sample_weight
from sklearn.linear_model import LogisticRegressionCV, LogisticRegression

import lightgbm as lgb

from bayes_opt import BayesianOptimization
from bayes_opt.observer import JSONLogger
from bayes_opt.event import Events
from bayes_opt.util import load_logs

from .base_utils import save, timer, neg_log_loss, split_X_y, load
from .data_visualization import Echart

import os
import warnings

warnings.filterwarnings('ignore')

# TODO
# 对逻辑回归调参添加随机/贝叶斯模式.

class CVStabilityEvaluator:
    '''
    CV稳定性验证.
    '''

    def __init__(self, train=None, test=None, label='target',
                 nfold=7, n_iter=50, save_path='./image/',
                 test_eval_metric='logloss', is_rank_corr_plot=False,
                 fig_title='1', plot_mode='html'):
        self.train = train
        self.test = test
        self.label = label
        self.nfold = nfold
        self.n_iter = n_iter
        self.test_eval_metric = test_eval_metric
        self.is_rank_corr_plot = is_rank_corr_plot
        self.fig_title = fig_title
        self.echart = Echart(mode=plot_mode)
        self.mean_list = []
        self.std_list = []
        self.test_result = []
        self.rank_corr_list = []
        self.params = {'objective': 'binary',
                       'num_tree': 1000,
                       'learning_rate': 0.05,
                       'max_depth': -1,
                       'num_leaf': 8,
                       'bagging_freq': 1,
                       'bagging_fraction': 0.6,
                       'feature_fraction': 0.4,
                       'min_data_in_leaf': 3,
                       'min_sum_hessian_in_leaf': 0.005,
                       'lambda_l1': 0.,
                       'lambda_l2': 0.,
                       'min_split_gain': 0.,
                       # 'is_unbalance': True,
                       'metrics': 'binary_logloss',
                       'seed': 7,
                       'num_thread': 4,
                       'verbose': -1
                       }
        self.save_path = save_path
        if not os.path.exists(save_path):
            os.mkdir(save_path)

    def random_set_params(self, seed=7):
        np.random.seed(seed)
        self.params['num_leaf'] = np.random.randint(2, 16, 1)[0]
        self.params['bagging_fraction'] = np.random.rand(1)[0]
        self.params['feature_fraction'] = np.random.rand(1)[0]
        self.params['min_data_in_leaf'] = np.random.randint(2, 50, 1)[0]
        self.params['min_sum_hessian_in_leaf'] = np.random.rand(1)[0] * 0.01
        self.params['lambda_l1'] = np.random.rand(1)[0] * 0.01
        self.params['lambda_l2'] = np.random.rand(1)[0] * 0.01

    @timer
    def run(self):
        '''CV稳定性测试 '''
        # 划分数据.
        if self.test is None:
            train, test = train_test_split(self.train, test_size=0.3, random_state=7, stratify=self.train[self.label])
            train_X, train_y = split_X_y(df=train, label=self.label)
            test_X, test_y = split_X_y(df=test, label=self.label)
        else:
            train_X, train_y = split_X_y(df=self.train, label=self.label)
            test_X, test_y = split_X_y(df=self.test, label=self.label)
        # 一阶段得到不同超参数下的CV均值和标准差.
        weight = compute_sample_weight(class_weight='balanced', y=train_y)
        train = lgb.Dataset(data=train_X, label=train_y, free_raw_data=False, weight=weight)
        for n in range(self.n_iter):
            print('第%d轮CV...' % (n + 1))
            self.random_set_params(seed=n)
            cv_result = lgb.cv(params=self.params, train_set=train, nfold=self.nfold,
                               early_stopping_rounds=30, verbose_eval=0)
            model = lgb.train(params=self.params, train_set=train,
                              num_boost_round=len(cv_result['binary_logloss-mean']))

            self.mean_list.append(cv_result['binary_logloss-mean'][-1])
            self.std_list.append(cv_result['binary_logloss-stdv'][-1])
            if self.test_eval_metric == 'logloss':
                self.test_result.append(-neg_log_loss(test_y, model.predict(test_X)))
            elif self.test_eval_metric == 'auc':
                self.test_result.append(roc_auc_score(test_y, model.predict(test_X)))

        # 画图展示.

        # CV-test结果关系图
        x_y = sorted(zip(self.mean_list, self.test_result), key=lambda x: x[0])
        x = [x[0] for x in x_y]
        y = [x[1] for x in x_y]

        self.echart.line(x=x, y=y, title=self.fig_title + '_CV-test结果关系图', is_max_show=True,
                         x_name='CV result', y_name='test result')

        if self.is_rank_corr_plot:
            # 二阶段根据CV均值和标准差的组合, 提高秩相关系数.
            print('开始计算秩相关系数.')
            for alpha in range(50):
                alpha /= 100
                mean_std_combine = np.array(self.mean_list) + alpha * np.array(self.std_list)
                self.rank_corr_list.append(spearmanr(mean_std_combine, np.array(self.test_result))[0])

            # 秩相关系数-标准差系数图.
            self.echart.line(x=list(range(len(self.rank_corr_list))), y=self.rank_corr_list,
                             title=self.fig_title + '_秩相关系数-标准差系数图', is_max_show=True,
                             x_name='标准差权重', y_name='秩相关系数')


class CVStabilityEvaluatorWithLR:
    '''
    CV稳定性验证.
    '''

    def __init__(self, train=None, test=None, label='target',
                 nfold=10, n_iter=50, save_path='./image/'):
        self.train = train
        self.test = test
        self.label = label
        self.nfold = nfold
        self.n_iter = n_iter
        self.mean_list = []
        self.test_result = []
        self.save_path = save_path
        if not os.path.exists(save_path):
            os.mkdir(save_path)

    def random_set_params(self, seed=7):
        np.random.seed(seed)
        C = (np.random.rand() * 2 - 1) * 5
        return pow(10, C)

    @timer
    def eval(self):
        '''CV稳定性测试 '''
        # 划分数据.
        if self.test is None:
            train, test = train_test_split(self.train, test_size=0.3, random_state=7, stratify=self.train[self.label])
            train_X, train_y = split_X_y(df=train, label=self.label)
            test_X, test_y = split_X_y(df=test, label=self.label)
        else:
            train_X, train_y = split_X_y(df=self.train, label=self.label)
            test_X, test_y = split_X_y(df=self.test, label=self.label)
        # 一阶段得到不同超参数下的CV均值.
        for n in range(self.n_iter):
            print('第%d轮CV...' % (n + 1))
            model = LogisticRegressionCV(Cs=[self.random_set_params(seed=n)],
                                         cv=self.nfold,
                                         penalty='l2',
                                         scoring='neg_log_loss',
                                         solver='liblinear',
                                         class_weight='balanced',
                                         max_iter=100,
                                         n_jobs=4,
                                         random_state=777,
                                         verbose=0
                                         )
            model.fit(train_X, train_y)
            cv_result = np.mean(model.scores_[1], axis=0).max()
            self.mean_list.append(cv_result)
            self.test_result.append(neg_log_loss(test_y, model.predict_proba(test_X)[:, 1]))

        # 画图展示.

        # CV-test结果关系图
        plt.figure(figsize=(16, 12))
        x_y = sorted(zip(self.mean_list, self.test_result), key=lambda x: x[0])
        x = [x[0] for x in x_y]
        y = [x[1] for x in x_y]
        plt.plot(x, y)
        plt.savefig(self.save_path + 'CV-test结果关系图')
        plt.show()


class LightCV:
    '''
    LGB CV.
    按层生长, 用于特征工程验证方法是否有效.
    '''

    def __init__(self,
                 df=None,
                 label='target',
                 nfold=7,
                 params=None,
                 eval_metric='logloss'):
        self.df = df
        self.label = label
        self.ft_list = self.df.columns.tolist()
        if self.label in self.ft_list:
            self.ft_list.remove(self.label)
        else:
            print('警告, 数据集中无目标列.')
        self.weight = compute_sample_weight(class_weight='balanced', y=self.df[self.label])
        self.train_set = lgb.Dataset(data=self.df[self.ft_list], label=self.df[self.label],
                                     free_raw_data=False, weight=self.weight)

        self.nfold = nfold
        if params is None:
            self.params = {'objective': 'binary',
                           'num_tree': 1000,
                           'learning_rate': 0.05,
                           'max_depth': 3,
                           'num_leaf': 256,
                           'bagging_freq': 1,
                           'bagging_fraction': 0.6,
                           'feature_fraction': 0.4,
                           'min_data_in_leaf': 3,
                           'min_sum_hessian_in_leaf': 0.005,
                           'lambda_l1': 0.,
                           'lambda_l2': 0.,
                           'min_split_gain': 0.,
                           'metrics': 'binary_logloss',
                           'seed': 7,
                           'num_thread': 4,
                           'verbose': -1
                           }
        # 保存路径.
        elif isinstance(params, str):
            self.params = load(params)
        # 参数字典对象.
        else:
            self.params = params
        if eval_metric == 'logloss':
            self.params['metrics'] = 'binary_logloss'
        elif eval_metric == 'auc':
            self.params['metrics'] = 'auc'
        else:
            print('警告, 评估指标错误.')
        self.eval_metric = eval_metric

    def run(self):
        cv_result = lgb.cv(params=self.params, train_set=self.train_set, nfold=self.nfold,
                           early_stopping_rounds=50)
        if self.eval_metric == 'logloss':
            return -cv_result['binary_logloss-mean'][-1], len(cv_result['binary_logloss-mean'])
        elif self.eval_metric == 'auc':
            return cv_result['auc-mean'][-1], len(cv_result['auc-mean'])


class LightBayesOptimizer:
    '''
    lightgbm二分类问题的贝叶斯优化.
    可以获得CV上的最佳超参数, 以及最佳模型.
    在创建对象的时候, 需要指定优化模式.
    1. 利用CV还是HO进行优化.
    2. 利用AUC还是LOGLOSS进行优化.
    '''

    def __init__(self, df=None, label=None, pbounds=None,
                 cv_verbose=-1, time_filed=None, val_size=0.3,
                 nfold=7, opt_mode='cv', metrics='logloss'):
        '''
        :param df: df, 数据集, 仅包含特征和目标列.
        :param label: str, 目标列名称.
        :param pbounds: dict, 超参数调参范围.
        :param cv_verbose: int, CV信息输出模式, 默认-1(不输出信息).
        :param time_filed: str, 时间字段名称.
        :param val_size: float, 验证集大小比例, 默认0.3.
        :param nfold: int, CV折数, 默认5.
        :param opt_mode: str, 优化模式, 可选'cv', 'ho', 默认'cv'
        :param metrics: str, 模型评估参数, 可选'auc', 'logloss', 默认'logloss'.
        '''
        target_data = df[label]
        ft_list = df.columns.tolist()
        ft_list.remove(label)
        if time_filed:
            ft_list.remove(time_filed)
        weight = compute_sample_weight(class_weight='balanced', y=df[label])
        self.df = lgb.Dataset(data=df[ft_list], label=target_data, free_raw_data=False, weight=weight)
        if opt_mode == 'cv':
            pass
        else:
            if time_filed:
                df = df.sort_values(time_filed)
                train = df.iloc[:int(len(df) * (1 - val_size))]
                val = df.iloc[int(len(df) * (1 - val_size)):]
                weight = compute_sample_weight(class_weight='balanced', y=train[label])
                self.train = lgb.Dataset(data=train[ft_list], label=train[label], free_raw_data=False, weight=weight)
                weight = compute_sample_weight(class_weight='balanced', y=val[label])
                self.val = lgb.Dataset(data=val[ft_list], label=val[label], free_raw_data=False, weight=weight)

            else:
                train, val = train_test_split(df, test_size=val_size,
                                              random_state=777, shuffle=True,
                                              stratify=df[label])

                weight = compute_sample_weight(class_weight='balanced', y=train[label])
                self.train = lgb.Dataset(data=train[ft_list], label=train[label], free_raw_data=False, weight=weight)
                weight = compute_sample_weight(class_weight='balanced', y=val[label])
                self.val = lgb.Dataset(data=val[ft_list], label=val[label], free_raw_data=False, weight=weight)
        if metrics == 'auc':
            self.metrics = 'auc'
        else:
            self.metrics = 'binary_logloss'
        self.opt_mode = opt_mode
        self.params = {'objective': 'binary',
                       'num_tree': 1000,
                       'learning_rate': 0.01,
                       'max_depth': -1,
                       'num_leaf': 8,
                       'bagging_freq': 1,
                       'bagging_fraction': 0.8,
                       'feature_fraction': 0.8,
                       'min_data_in_leaf': 3,
                       'min_sum_hessian_in_leaf': 0.005,
                       'lambda_l1': 0.,
                       'lambda_l2': 0.,
                       'min_split_gain': 0.,
                       # 'is_unbalance': True,
                       'metrics': self.metrics,
                       'seed': 7,
                       'num_thread': 4,
                       'verbose': -1
                       }

        # 设置叶斯优化的超参数范围.
        self.pbounds = {'num_leaf': (2, 42),
                        'bagging_fraction': (0.2, 1),
                        'feature_fraction': (0.2, 1),
                        'min_data_in_leaf': (1, 50),
                        'min_sum_hessian_in_leaf': (-5, 1),
                        'lambda_l1': (-5, 0),
                        'lambda_l2': (-5, 0),
                        'min_split_gain': (-5, 0)}
        if pbounds:
            for key in pbounds:
                self.pbounds[key] = pbounds[key]
        self.val_size = val_size
        self.nfold = nfold
        self.cv_verbose = cv_verbose

        self.model = None
        self.best_params = None
        self.best_result = -np.inf

    def cv(self, **kwargs):
        '''
        贝叶斯优化目标函数.
        通过传入不同的参数, 返回对应的函数值.
        '''
        params = self.params.copy()
        params['num_tree'] = 10000
        params['num_leaf'] = int(kwargs['num_leaf'])
        params['bagging_fraction'] = kwargs['bagging_fraction']
        params['feature_fraction'] = kwargs['feature_fraction']
        params['min_data_in_leaf'] = int(kwargs['min_data_in_leaf'])
        params['min_sum_hessian_in_leaf'] = pow(10, kwargs['min_sum_hessian_in_leaf']) - 1e-5
        params['lambda_l1'] = pow(10, kwargs['lambda_l1']) - 1e-5
        params['lambda_l2'] = pow(10, kwargs['lambda_l2']) - 1e-5
        params['min_split_gain'] = pow(10, kwargs['min_split_gain']) - 1e-5

        cv_result = lgb.cv(params=params, train_set=self.df, nfold=self.nfold,
                           early_stopping_rounds=50, verbose_eval=self.cv_verbose)
        if self.metrics == 'binary_logloss':
            result = -cv_result['binary_logloss-mean'][-1]
            num_tree = len(cv_result['binary_logloss-mean'])
        else:
            result = cv_result['auc-mean'][-1]
            num_tree = len(cv_result['auc-mean'])
        if result > self.best_result:
            self.best_params = params
            self.best_params['num_tree'] = num_tree
            self.best_result = result
        return result

    def ho(self, **kwargs):
        params = self.params.copy()
        params['num_tree'] = 10000
        params['num_leaf'] = int(kwargs['num_leaf'])
        params['bagging_fraction'] = kwargs['bagging_fraction']
        params['feature_fraction'] = kwargs['feature_fraction']
        params['min_data_in_leaf'] = int(kwargs['min_data_in_leaf'])
        params['min_sum_hessian_in_leaf'] = pow(10, kwargs['min_sum_hessian_in_leaf']) - 1e-5
        params['lambda_l1'] = pow(10, kwargs['lambda_l1']) - 1e-5
        params['lambda_l2'] = pow(10, kwargs['lambda_l2']) - 1e-5
        params['min_split_gain'] = pow(10, kwargs['min_split_gain']) - 1e-5

        model = lgb.train(params=params, train_set=self.train,
                          num_boost_round=10000, valid_sets=[self.val], valid_names=['val'],
                          early_stopping_rounds=50, verbose_eval=False)
        if self.metrics == 'binary_logloss':
            result = -model.best_score['val']['binary_logloss']
        else:
            result = model.best_score['val']['auc']
        if result > self.best_result:
            self.best_params = params
            self.best_params['num_tree'] = model.best_iteration
            self.best_result = result

        return result

    @timer
    def run(self, save_log_path='bayes_opt_log.json',
            if_save_log=False, if_load_log=False, init_points=50,
            n_iter=0, verbose=1, is_verbose=1, random_state=777):
        '''
        lgb贝叶斯优化主函数.
        :param save_log_path: log保存地址, str.
        :param if_save_log: 是否保存log.
        :param if_load_log: 是否加载log.
        :return: 贝叶斯优化结果.
        '''
        if self.opt_mode == 'ho':
            f = self.ho
        else:
            f = self.cv
        opt = BayesianOptimization(f=f, pbounds=self.pbounds,
                                   random_state=random_state, verbose=verbose)

        # 加载log.
        if if_load_log and os.path.exists(save_log_path):
            load_logs(opt, logs=[save_log_path])

        # 保存log.
        if if_save_log:
            logger = JSONLogger(path=save_log_path)
            opt.subscribe(Events.OPTMIZATION_STEP, logger)

        # 贝叶斯搜索.
        opt.maximize(init_points=init_points, n_iter=n_iter)

        # 打印最佳结果.
        if is_verbose:
            print('Best result is:')
            print(opt.max['target'])

        return opt.max['target']

    def fit_model(self, do_correct=True):
        params = self.best_params.copy()
        if do_correct:
            if self.opt_mode == 'ho':
                correct_ratio = 1 / (1 - self.val_size)
            else:
                correct_ratio = 1 / (1 - 1 / self.nfold)
            params['min_data_in_leaf'] = int(params['min_data_in_leaf'] * correct_ratio)
            params['min_sum_hessian_in_leaf'] *= correct_ratio
            params['lambda_l1'] *= correct_ratio
            params['lambda_l2'] *= correct_ratio
            params['min_split_gain'] *= correct_ratio
        self.model = lgb.train(params=params, train_set=self.df)

    def predict(self, data):
        return self.model.predict(data[self.model.feature_name()])

    def save_model(self, path='./model/lgb.model'):
        save(path=path, obj=self.model)

    def save_params(self, path='./model/lgb.params'):
        save(path=path, obj=self.best_params)


class LightBayesOptimizerRough:
    '''
    lightgbm二分类问题的贝叶斯粗优化, 主要用于特征工程中.
    '''

    def __init__(self, df=None, label=None, pbounds=None,
                 cv_verbose=-1, nfold=7):
        '''
        :param df: df, 数据集, 仅包含特征和目标列.
        :param label: str, 目标列名称.
        :param pbounds: dict, 超参数调参范围.
        :param cv_verbose: int, CV信息输出模式, 默认-1(不输出信息).
        :param nfold: int, CV折数, 默认5.
        '''
        target_data = df[label]
        ft_list = df.columns.tolist()
        ft_list.remove(label)
        weight = compute_sample_weight(class_weight='balanced', y=df[label])
        self.df = lgb.Dataset(data=df[ft_list], label=target_data, free_raw_data=False, weight=weight)
        self.params = {'objective': 'binary',
                       'num_tree': 1000,
                       'learning_rate': 0.05,
                       'max_depth': 3,
                       'num_leaf': 256,
                       'bagging_freq': 1,
                       'bagging_fraction': 0.6,
                       'feature_fraction': 0.4,
                       'min_data_in_leaf': 3,
                       'min_sum_hessian_in_leaf': 0.005,
                       'lambda_l1': 0.,
                       'lambda_l2': 0.,
                       'min_split_gain': 0.,
                       'metrics': 'binary_logloss',
                       'seed': 7,
                       'num_thread': 4,
                       'verbose': -1
                       }

        # 设置叶斯优化的超参数范围.
        self.pbounds = {'max_depth': (1, 6),
                        'bagging_fraction': (0.2, 1),
                        'feature_fraction': (0.2, 1),
                        'min_sum_hessian_in_leaf': (-5, 2)}
        if pbounds:
            for key in pbounds:
                self.pbounds[key] = pbounds[key]
        self.nfold = nfold
        self.cv_verbose = cv_verbose

        self.model = None
        self.best_params = None
        self.best_result = -np.inf

    def cv(self, **kwargs):
        '''
        贝叶斯优化目标函数.
        通过传入不同的参数, 返回对应的函数值.
        '''
        params = self.params.copy()
        params['num_tree'] = 10000
        params['max_depth'] = int(kwargs['max_depth'])
        params['bagging_fraction'] = kwargs['bagging_fraction']
        params['feature_fraction'] = kwargs['feature_fraction']
        params['min_sum_hessian_in_leaf'] = pow(10, kwargs['min_sum_hessian_in_leaf']) - 1e-5

        cv_result = lgb.cv(params=params, train_set=self.df, nfold=self.nfold,
                           early_stopping_rounds=50, verbose_eval=self.cv_verbose)
        result = -cv_result['binary_logloss-mean'][-1]
        num_tree = len(cv_result['binary_logloss-mean'])
        if result > self.best_result:
            self.best_params = params
            self.best_params['num_tree'] = num_tree
            self.best_result = result
        return result

    @timer
    def run(self, save_log_path='bayes_opt_log.json',
            if_save_log=False, if_load_log=False, init_points=30,
            n_iter=10, verbose=0, is_verbose=1, random_state=7):
        '''贝叶斯粗优化 '''
        opt = BayesianOptimization(f=self.cv, pbounds=self.pbounds,
                                   random_state=random_state, verbose=verbose)

        # 加载log.
        if if_load_log and os.path.exists(save_log_path):
            load_logs(opt, logs=[save_log_path])

        # 保存log.
        if if_save_log:
            logger = JSONLogger(path=save_log_path)
            opt.subscribe(Events.OPTMIZATION_STEP, logger)

        # 贝叶斯搜索.
        opt.maximize(init_points=init_points, n_iter=n_iter)

        # 打印最佳结果.
        if is_verbose:
            print('Best result is:')
            print(opt.max['target'])

        return opt.max['target']

    def fit_model(self, do_correct=True):
        params = self.best_params.copy()
        if do_correct:
            correct_ratio = 1 / (1 - 1 / self.nfold)
            params['min_sum_hessian_in_leaf'] *= correct_ratio
        self.model = lgb.train(params=params, train_set=self.df)

    def predict(self, data):
        return self.model.predict(data[self.model.feature_name()])

    def save_model(self, path='./model/lgb.model'):
        save(path=path, obj=self.model)

    def save_params(self, path='./model/lgb.params'):
        save(path=path, obj=self.best_params)


class LRCV:
    '''
    逻辑回归CV.
    '''

    def __init__(self, nfold=7, Cs=10, penalty='l2', metric='logloss'):
        if metric == 'logloss':
            scoring = 'neg_log_loss'
        elif metric == 'auc':
            scoring = self.auc
        else:
            print('警告, 评估参数输入错误.')

        self.model = LogisticRegressionCV(Cs=Cs,
                                          cv=nfold,
                                          penalty=penalty,
                                          scoring=scoring,
                                          solver='liblinear',
                                          class_weight='balanced',
                                          max_iter=100,
                                          n_jobs=4,
                                          random_state=7,
                                          verbose=0)

    def cv(self, X=None, y=None):
        self.model.fit(X, y)
        return np.mean(self.model.scores_[1], axis=0).max()

    def auc(self, estimator, X, y):
        y_pred = estimator.predict_proba(X)[:, 1]
        return roc_auc_score(y_true=y, y_score=y_pred)


class LROptimizer:
    '''
    二分类逻辑回归参数优化.
    '''

    def __init__(self, df=None, label='target',
                 nfold=7, Cs=10, val_perc=0.3, mode='cv',
                 class_weight='balanced'
                 ):
        self.label = label
        self.ft_list = df.columns.tolist()
        if label in self.ft_list:
            self.ft_list.remove(label)
        else:
            print('警告, 数据集中不包含目标列.')
        self.nfold = nfold
        self.Cs = Cs
        self.model = None
        self.df = df
        self.mode = mode
        self.val_perc = val_perc
        self.class_weight = class_weight

        # 若为hold out模式, 则划分出验证集.
        if self.mode == 'ho':
            self.train, self.val = train_test_split(self.df,
                                                    test_size=self.val_perc,
                                                    random_state=7,
                                                    stratify=self.df[self.label])

    @timer
    def run(self):
        '''逻辑回归参数调节 '''
        print('正在训练逻辑回归模型.')
        if self.mode == 'cv':
            # L1.
            lr_model_l1 = LogisticRegressionCV(Cs=self.Cs,
                                               cv=self.nfold,
                                               penalty='l1',
                                               scoring='neg_log_loss',
                                               solver='liblinear',
                                               class_weight=self.class_weight,
                                               max_iter=100,
                                               n_jobs=4,
                                               random_state=7,
                                               verbose=0)
            lr_model_l1.fit(self.df[self.ft_list], self.df[self.label])
            score_l1 = np.mean(lr_model_l1.scores_[1], axis=0).max()

            # L2
            lr_model_l2 = LogisticRegressionCV(Cs=self.Cs,
                                               cv=self.nfold,
                                               penalty='l2',
                                               scoring='neg_log_loss',
                                               solver='liblinear',
                                               class_weight=self.class_weight,
                                               max_iter=100,
                                               n_jobs=4,
                                               random_state=7,
                                               verbose=0)
            lr_model_l2.fit(self.df[self.ft_list], self.df[self.label])
            score_l2 = np.mean(lr_model_l2.scores_[1], axis=0).max()

            if score_l1 > score_l2:
                self.model = lr_model_l1
            else:
                self.model = lr_model_l2
            self.C = self.model.C_[0]
            self.penalty = self.model.get_params()['penalty']
            print('最佳模型的参数:')
            print('正则系数: %f, 正则模式: %s' % (self.C, self.penalty))
        else:
            for C in [0.0001, 0.001, 0.01, 0.1, 1, 10, 100, 1000, 10000]:
                lr_model_l1 = LogisticRegression(penalty='l1',
                                                 C=C,
                                                 class_weight=self.class_weight,
                                                 random_state=7,
                                                 solver='liblinear',
                                                 max_iter=100,
                                                 n_jobs=4)
                lr_model_l1.fit(self.train[self.ft_list], self.train[self.label])

                score_l1 = neg_log_loss(self.val[self.label], lr_model_l1.predict_proba(self.val[self.ft_list]))

                lr_model_l2 = LogisticRegression(penalty='l2',
                                                 C=C,
                                                 class_weight=self.class_weight,
                                                 random_state=7,
                                                 solver='liblinear',
                                                 max_iter=100,
                                                 n_jobs=4)
                lr_model_l2.fit(self.train[self.ft_list], self.train[self.label])
                score_l2 = neg_log_loss(self.val[self.label], lr_model_l2.predict_proba(self.val[self.ft_list]))

                if score_l1 > score_l2:
                    self.model = lr_model_l1
                else:
                    self.model = lr_model_l2
                self.C = self.model.C
                self.penalty = self.model.get_params()['penalty']
            print('最佳模型的参数:')
            print('正则系数: %f, 正则模式: %s' % (self.C, self.penalty))

    def predict(self, data):
        data = data[self.ft_list]
        return self.model.predict_proba(data)[:, 1]

    def save_model(self, path='./model/LR.model'):
        save(path=path, obj=self.model)


if __name__ == '__main__':
    from base_utils import split_X_y

    train = pd.read_csv('./data/tw_train.csv')
    test = pd.read_csv('./data/tw_test.csv')
    df = train.append(test)
    df_X, df_y = split_X_y(df)
    cvse = CVStabilityEvaluator(nfold=10, n_iter=50)
    cvse.run(df_X, df_y)



# coding=utf-8
'''
连接数据库脚本.
'''
import pymysql
import pymongo

import pandas as pd

class MySQLConnector:
    def __init__(self):
        self.mysql_engine = pymysql.connect(host='172.20.6.9',
                                            port=9030,
                                            user='fengkong_read_only',
                                            passwd='mT2HFUgI',
                                            db='risk_analysis',
                                            charset='utf8')
        pass


    def query(self, sql):
        try:
            return pd.read_sql(sql, self.mysql_engine)
        except:
            print('SQL查询出现错误.')
            return None



class MongoConnector:
    def __init__(self, mode='all'):
        # 审核+空跑
        if mode == 'all':
            table_name = 'wf_audit_log_with_feature'
        # 审核
        elif mode == 'online':
            table_name = 'wf_audit_log_with_feature_online'
        # 空跑
        elif mode == 'offline':
            table_name = 'wf_audit_log_with_feature_offline'
        else:
            print('警告, 模式设置错误.')

        self.mongo_client = pymongo.MongoClient(
            "mongodb://haoyue.shu:x2egwRHk7WhQ4So1@172.18.3.22:27017/?authSource=rc_mgo_feature_dp")
        self.mongo_db = self.mongo_client['rc_mgo_feature_dp']
        self.mongo_table = self.mongo_db[table_name]


    def query(self, condition, fields):
        try:
            return pd.DataFrame(list(self.mongo_table.find(condition, fields)))
        except:
            print('Mongo查询出现错误.')


if __name__ == '__main__':
    # MySQL查询.
    field_list = \
        '''
        'order_no', 'loan_id', 'user_id', 'uuid',
        'applied_from', 'applied_channel',
        'applied_at', 'deadline',  'passdue_day',
        'term_no', 
        'transacted', 
        'reloan_v3_point'
        '''
    field_list = field_list.replace('\n', ' ').replace('\'', ' ')
    mysql_con = MySQLConnector()
    mysql_df = mysql_con.query('SELECT %s \
                         FROM risk_analysis \
                         WHERE  transacted=1 \
                         AND applied_at >= "2019-05-10 00:00:00" \
                         AND applied_at <= "2019-07-10 00:00:00"' % field_list)


    # Mongo查询.
    # mongo_db中可能需要的字段。

    need_cols = {'order_id': 1,  # 订单号
                 'wf_loan_type': 1,  # 申请类型
                 'wf_created_at': 1,  # 申请时间
                 'wf_biz_channel': 1,  # 申请渠道号
                 'passdue_day': 1,  # 逾期时间
                 'user_uuid': 1,  # uuid
                 'lam_transaction_status': 1,  # 2和5表示放款成功.
                 'repayment_status': 1, # 不等于4表示放款成功.
                 'model_exec_data_source#fst_v6_xy_br_dhb_raw': 1,
                 'model_exec_data_source#fst_withoutoperator_v1': 1,
                 'model_exec_data_source#fst_v6_xy_br_dhb_tencent_contact_raw': 1,
                 'model_exec_data_source#reloan_v6_xy_tz_bj': 1,
                 'model_exec_data_source#reloan_xy_tz_bj': 1,
                 'model_exec_data_source#bairong_raw': 1,
                 'model_exec_data_source#bairong_v2_raw': 1,
                 'model_exec_data_source#dhb_score': 1,
                 'model_exec_data_source#dhb_v2': 1,
                 'model_exec_data_source#tanzhi_score': 1,
                 'lxf_score_scale_v2': 1,
                 'model_exec_data_source#reloan_v4_raw': 1,
                 'model_exec_data_source#xinyan_v2': 1,
                 'model_exec_data_source#xinyan_v3': 1,
                 'model_exec_data_source#moxie_raw': 1,
                 'model_exec_data_source#tongdun_v2': 1,
                 'model_exec_data_source#operator_score_raw': 1,
                 'tongdunScore': 1,
                 'third_data_source#qcaf_af_rscore': 1,
                 'third_data_source#bj_hawk_bj_score': 1
                 }
    mongo_con = MongoConnector(mode='all')
    condition = {'wf_created_at': {'$gte': '%s 00:00:00' % '2019-01-01',
                                   '$lte': '%s 00:00:00' % '2019-02-01'},
                 'passdue_day': {'$ne': None}}
    mongo_df = mongo_con.query(condition, need_cols)
    pass


# coding=utf-8
'''
特征选择.
'''
import numpy as np
import pandas as pd
from collections import OrderedDict
from sklearn.model_selection import train_test_split, StratifiedKFold
from sklearn.utils.class_weight import compute_sample_weight
from sklearn.linear_model import LogisticRegressionCV
from sklearn.metrics import log_loss

import lightgbm as lgb
from eli5.permutation_importance import get_score_importances
from eli5.sklearn.permutation_importance import PermutationImportance
from .base_class import BorutaPy
from .base_utils import timer, rank_sort_combine, save, traversal_search, neg_log_loss, load
from .param_optimize import LightCV, LRCV, LightBayesOptimizerRough
from .data_transform import StandardTransformer
from .data_visualization import Echart

import matplotlib.pyplot as plt
import seaborn as sns

import pickle
import warnings

warnings.filterwarnings('ignore')


class ModelParamsGenerator:
    '''
    利用重要性, 得到不同特征数量对应的模型参数.
    '''

    def __init__(self, df=None, label='target',
                 imp_type='split',
                 ft_percs=[20, 40, 60, 80, 100],
                 ft_percs_range=[(0., .3), (.3, .5), (.5, .7), (.7, .9), (.9, 1.)],
                 nfolds=7,
                 init_points=20,
                 n_iter=10,
                 save_path='ft_perc_range_params.dict'):
        self.label = label
        self.df = df
        self.n_ft = self.df.shape[1] - 1
        self.nfolds = nfolds
        self.init_points = init_points
        self.n_iter = n_iter
        self.imp_type = imp_type
        self.ft_percs = ft_percs
        self.ft_percs.reverse()
        self.ft_percs_range = ft_percs_range
        self.ft_percs_range.reverse()
        self.save_path = save_path

        self.ft_list = df.columns.tolist()
        if label in self.ft_list:
            self.ft_list.remove(label)
        else:
            print('警告, 数据中没有目标列')

        self.ft_params_dict = OrderedDict()

    @timer
    def run(self):
        '''获取不同特征数量对应模型参数 '''
        tmp_ft_list = self.ft_list.copy()

        for i in range(len(self.ft_percs)):
            print('第%d轮, 特征数量%d' % (i + 1, len(tmp_ft_list)))
            # 跑CV得到一组相对合适的超参数.
            lbo = LightBayesOptimizerRough(df=self.df[tmp_ft_list + [self.label]], label=self.label, nfold=self.nfolds)
            lbo.run(init_points=self.init_points, n_iter=self.n_iter, verbose=0)
            lbo.fit_model()
            self.ft_params_dict[self.ft_percs_range[i]] = lbo.best_params

            # 利用该超参数训练模型并筛选特征.
            imp_list = lbo.model.feature_importance(importance_type=self.imp_type)
            ft_imp_list = list(zip(tmp_ft_list, imp_list))
            ft_imp_list = sorted(ft_imp_list, key=lambda x: x[1], reverse=True)
            if i + 1 < len(self.ft_percs):
                tmp_ft_list = [x[0] for x in ft_imp_list][: int(self.ft_percs[i + 1] / 100 * self.n_ft)]

        save(path='./model/' + self.save_path, obj=self.ft_params_dict)
        print('运行完成, 保存模型参数成功.')
        return self.ft_params_dict


class LGBRFE:
    def __init__(self, df=None, label='target',
                 nfold=7,
                 imp_type='split',
                 min_num_ft=0.1,
                 n_ft_rm_per_iter=0.05,
                 params=None):
        self.df = df
        self.n_sample = self.df.shape[0]
        self.label = label
        self.imp_type = imp_type
        self.nfold = nfold
        self.ft_list = self.df.columns.tolist()
        if self.label in self.ft_list:
            self.ft_list.remove(self.label)
        else:
            print('警告, 数据中不包含目标列.')

        self.min_num_ft = min_num_ft if isinstance(min_num_ft, int) \
            else int(self.df.shape[1] * min_num_ft)
        self.n_ft_rm_per_iter = n_ft_rm_per_iter if isinstance(n_ft_rm_per_iter, int) \
            else int(self.df.shape[1] * n_ft_rm_per_iter)
        if self.n_ft_rm_per_iter == 0:
            self.n_ft_rm_per_iter = 1

        self.rfe_result = []
        self.best_ft_list = None
        self.best_result = None
        self.weight = compute_sample_weight(class_weight='balanced', y=self.df[self.label])

        if params is None:
            self.params = {'objective': 'binary',
                           'num_tree': 1000,
                           'learning_rate': 0.05,
                           'max_depth': 3,
                           'num_leaf': 64,
                           'bagging_freq': 1,
                           'bagging_fraction': 0.6,
                           'feature_fraction': 0.4,
                           'min_data_in_leaf': 3,
                           'min_sum_hessian_in_leaf': 0.005,
                           'lambda_l1': 0.,
                           'lambda_l2': 0.,
                           'min_split_gain': 0.,
                           'metrics': 'binary_logloss',
                           'seed': 7,
                           'num_thread': 4,
                           'verbose': -1
                           }
        elif isinstance(params, str):
            self.params = load(params)
        else:
            self.params = params
        if not isinstance(list(self.params.keys())[0], tuple):
            self.params = {(0., 1.): self.params}

    def choose_params(self, ft_list):
        perc = len(ft_list) / self.n_sample
        for perc_range in self.params:
            if perc >= perc_range[0] and perc <= perc_range[1]:
                return self.params[perc_range], perc_range

    @timer
    def run(self):
        '''LGBRFE '''
        remain_ft_list = self.ft_list.copy()
        print('开始RFE特征选择...')
        cv = LightCV(df=self.df,
                     label=self.label,
                     nfold=self.nfold,
                     params=self.choose_params(remain_ft_list)[0])
        cv_result, n_tree = cv.run()
        self.rfe_result.insert(0, (len(remain_ft_list), cv_result, remain_ft_list))
        self.best_ft_list, self.best_result = remain_ft_list, cv_result
        print('第1轮, 特征数量%d, cv结果%f' % (len(remain_ft_list), cv_result))
        n_iter = 1
        while len(remain_ft_list) > self.min_num_ft:

            train_set = lgb.Dataset(data=self.df[remain_ft_list], label=self.df[self.label],
                                    free_raw_data=False, weight=self.weight)
            model = lgb.train(params=self.choose_params(remain_ft_list)[0], train_set=train_set)

            imp_list = model.feature_importance(importance_type=self.imp_type)
            ft_imp_list = sorted(zip(remain_ft_list, imp_list), key=lambda x: x[1], reverse=True)
            remain_ft_list = [x[0] for x in ft_imp_list]
            if len(remain_ft_list) > self.n_ft_rm_per_iter + self.min_num_ft:
                remain_ft_list = remain_ft_list[: -self.n_ft_rm_per_iter]
            else:
                remain_ft_list = remain_ft_list[: self.min_num_ft]
            cv = LightCV(df=self.df[remain_ft_list + [self.label]],
                         label=self.label,
                         nfold=self.nfold,
                         params=self.choose_params(remain_ft_list)[0])
            cv_result, n_tree = cv.run()
            self.rfe_result.insert(0, (len(remain_ft_list), cv_result, remain_ft_list))
            n_iter += 1
            self.best_result = cv_result if cv_result > self.best_result else self.best_result
            self.best_ft_list = remain_ft_list if cv_result == self.best_result else self.best_ft_list
            print('第%d轮, 特征数量%d, cv结果%f' % (n_iter, len(remain_ft_list), cv_result))
        print('最佳特征数量%d, 最佳CV结果%f' % (len(self.best_ft_list), self.best_result))

    def plot(self, mode='html', title='LGBRFE',
             x_name='feature nums', y_name='CV result'):

        X, Y = [], []
        for item in self.rfe_result:
            X.append(item[0])
            Y.append(item[1])
        echart = Echart(mode=mode)
        echart.line(x=X, y=Y,
                    title=title,
                    is_max_show=True,
                    x_name=x_name,
                    y_name=y_name)


class RFECV:
    def __init__(self, nfold=10,
                 rank_type='gain',
                 min_num_ft=0.1,
                 n_ft_rm_per_iter=0.05):
        self.min_num_ft = min_num_ft
        self.n_ft_rm_per_iter = n_ft_rm_per_iter
        self.cv = LightCV(nfold=nfold)
        self.rank_type = rank_type
        self.rfe_result = []
        self.model = lgb.LGBMClassifier(boosting_type='gbdt',
                                        num_leaves=64,
                                        max_depth=5,
                                        learning_rate=0.01,
                                        n_estimators=1000,
                                        class_weight='balanced',
                                        min_child_weight=0.001,
                                        min_child_samples=1,
                                        subsample=0.6,
                                        subsample_freq=1,
                                        colsample_bytree=1,
                                        random_state=7,
                                        n_jobs=4,
                                        importance_type='gain')
        self.imp_generator = None

    def fit_transform(self, X=None, y=None):
        self.min_num_ft = self.min_num_ft if isinstance(self.min_num_ft, int) \
            else int(X.shape[1] * self.min_num_ft)
        self.n_ft_rm_per_iter = self.n_ft_rm_per_iter if isinstance(self.n_ft_rm_per_iter, int) \
            else int(X.shape[1] * self.n_ft_rm_per_iter)
        remain_ft_list = X.columns.tolist()
        print('开始RFE特征选择...')
        cv_result = self.cv.run(X=X[remain_ft_list], y=y)
        self.rfe_result.insert(0, (len(remain_ft_list), cv_result, remain_ft_list))
        print('第1轮, 特征数量%d, cv结果%f' % (len(remain_ft_list), cv_result))
        n_iter = 1
        while len(remain_ft_list) > self.min_num_ft:
            self.model.set_params(n_estimators=int(100 * (len(remain_ft_list) /
                                                          (np.sqrt(len(remain_ft_list)) *
                                                           self.model.get_params()['max_depth']))),
                                  colsample_bytree=np.sqrt(len(remain_ft_list)) / len(remain_ft_list))
            if self.rank_type == 'gain':
                self.imp_generator = self.model
            elif self.rank_type == 'split':
                self.model.set_params(importance_type='split')
                self.imp_generator = self.model
            elif self.rank_type == 'eli5':
                self.imp_generator = PermutationImportance(estimator=self.model, scoring=self.neg_log_loss,
                                                           n_iter=5, random_state=7, cv=5, refit=False)
            elif self.rank_type == 'mix':
                self.imp_generator = PermutationImportance(estimator=self.model, scoring=self.neg_log_loss,
                                                           n_iter=5, random_state=7, cv=5, refit=True)
            else:
                print('不支持该排序类型.')
                return None

            self.imp_generator.fit(X, y, verbose=False)
            if self.rank_type != 'mix':
                imp_list = self.imp_generator.feature_importances_
            else:
                imp_list = rank_sort_combine([self.imp_generator.feature_importances_.tolist(),
                                              self.imp_generator.estimator_.feature_importances_.tolist()])
            ft_imp_list = sorted(zip(remain_ft_list, imp_list), key=lambda x: x[1], reverse=True)
            remain_ft_list = [x[0] for x in ft_imp_list]
            if len(remain_ft_list) > self.n_ft_rm_per_iter + self.min_num_ft:
                remain_ft_list = remain_ft_list[: -self.n_ft_rm_per_iter]
            else:
                remain_ft_list = remain_ft_list[: self.min_num_ft]
            cv_result = self.cv.run(X=X[remain_ft_list], y=y)
            self.rfe_result.insert(0, (len(remain_ft_list), cv_result, remain_ft_list))

            n_iter += 1
            print('第%d轮, 特征数量%d, cv结果%f' % (n_iter, len(remain_ft_list), cv_result))

    def neg_log_loss(self, model, X, y):
        y_pred = model.predict_proba(X)[:, 1]
        return neg_log_loss(y, y_pred)

    def plot(self):
        X, Y = [], []
        for item in self.rfe_result:
            X.append(item[0])
            Y.append(item[1])

        plt.figure(figsize=(16, 12))
        plt.plot(X, Y)
        plt.xlabel('feature nums')
        plt.ylabel('cv result')
        plt.savefig('rfe')
        plt.show()

        pass


class ForwardSelector:
    def __init__(self, df=None, label='target', nfold=7, Cs=10,
                 is_break=True, retain_ft_list=None):
        self.df = df.reset_index(drop=True)
        self.label = label
        self.is_break = is_break
        self.ft_list = df.columns.tolist()
        self.ft_list.remove(label)
        self.retain_ft_list = retain_ft_list
        self.model = LogisticRegressionCV(Cs=Cs, solver='liblinear', class_weight='balanced',
                                          n_jobs=4, random_state=7, cv=nfold, scoring='neg_log_loss')

        self.best_ft_list = None
        self.best_cv_result = -np.inf
        self.ft_result_dict = {}

    def cv_lr(self, ft_list):
        self.model.fit(self.df[ft_list], self.df[self.label])
        result = self.model.scores_
        return np.mean(result[1], axis=0).max()

    def run(self):
        if self.retain_ft_list is None:
            ft_result_list = []
            for ft in self.ft_list:
                ft_result_list.append((ft, self.cv_lr([ft])))
            ft_result_list = sorted(ft_result_list, key=lambda x: x[1], reverse=True)
            select_ft_list = [ft_result_list[0][0]]
            ft_list = self.ft_list.copy()
            ft_list.remove(select_ft_list[0])
            print('添加特征: ', ft_result_list[0][0], 'CV结果: ', ft_result_list[0][1])
            self.best_cv_result = ft_result_list[0][1]
            self.best_ft_list = [ft_result_list[0][0]]
            self.ft_result_dict[1] = (self.best_ft_list.copy(), self.best_cv_result)
        else:
            select_ft_list = self.retain_ft_list.copy()
            ft_list = self.ft_list.copy()
            for ft in select_ft_list:
                ft_list.remove(ft)
            self.best_cv_result = self.cv_lr(select_ft_list)
            self.best_ft_list = select_ft_list.copy()
            self.ft_result_dict[len(self.best_ft_list)] = (self.best_ft_list.copy(), self.best_cv_result)
            print('初始特征数量: ', len(self.best_ft_list), 'CV结果: ', self.best_ft_list)

        while len(ft_list) > 0:
            result_list = []
            for ft in ft_list:
                tmp_ft_list = select_ft_list + [ft]
                result = self.cv_lr(tmp_ft_list)
                result_list.append((ft, result))
            result_list = sorted(result_list, key=lambda x: x[1], reverse=True)

            if result_list[0][1] > self.best_cv_result:
                self.best_cv_result = result_list[0][1]
                select_ft_list.append(result_list[0][0])
                ft_list.remove(result_list[0][0])
                self.best_ft_list = select_ft_list.copy()
                print('添加特征: ', result_list[0][0], 'CV结果上升: ', result_list[0][1])
            else:
                if self.is_break:
                    break
                else:
                    select_ft_list.append(result_list[0][0])
                    ft_list.remove(result_list[0][0])
                    print('添加特征: ', result_list[0][0], 'CV结果下降: ', result_list[0][1])
            self.ft_result_dict[len(select_ft_list)] = (select_ft_list.copy(), result_list[0][1])
        return self.best_ft_list

    def save(self, save_path='./model/forward_ft.dict'):
        save(save_path, self.ft_result_dict)


class BackwardSelector:
    def __init__(self, df=None, label='target', nfold=7, Cs=10,
                 is_break=True, retain_ft_list=None):
        self.df = df.reset_index(drop=True)
        self.label = label
        self.is_break = is_break
        self.ft_list = df.columns.tolist()
        self.ft_list.remove(label)
        self.retain_ft_list = set(retain_ft_list) if retain_ft_list is not None else set()
        self.model = LogisticRegressionCV(Cs=Cs, solver='liblinear', class_weight='balanced',
                                          n_jobs=4, random_state=7, cv=nfold, scoring='neg_log_loss')

        self.best_ft_list = None
        self.best_cv_result = -np.inf
        self.ft_result_dict = {}

    def cv_lr(self, ft_list):
        self.model.fit(self.df[ft_list], self.df[self.label])
        result = self.model.scores_
        return np.mean(result[1], axis=0).max()

    def run(self):

        ft_list = self.ft_list.copy()
        self.best_cv_result = self.cv_lr(ft_list)
        self.best_ft_list = ft_list.copy()
        self.ft_result_dict[len(self.best_ft_list)] = (self.best_ft_list.copy(), self.best_cv_result)
        print('初始特征数量: ', len(self.best_ft_list), 'CV结果: ', self.best_ft_list)

        while len(ft_list) > len(self.retain_ft_list):
            result_list = []
            for ft in ft_list:
                tmp_ft_list = ft_list.copy()
                if ft not in self.retain_ft_list:
                    tmp_ft_list.remove(ft)
                else:
                    continue
                result = self.cv_lr(tmp_ft_list)
                result_list.append((ft, result))
            result_list = sorted(result_list, key=lambda x: x[1], reverse=True)

            if result_list[0][1] > self.best_cv_result:
                self.best_cv_result = result_list[0][1]
                ft_list.remove(result_list[0][0])
                self.best_ft_list = ft_list.copy()
                print('去掉特征: ', result_list[0][0], 'CV结果上升: ', result_list[0][1])
            else:
                if self.is_break:
                    break
                else:
                    ft_list.remove(result_list[0][0])
                    print('去掉特征: ', result_list[0][0], 'CV结果下降: ', result_list[0][1])
            self.ft_result_dict[len(ft_list)] = (ft_list.copy(), result_list[0][1])
        return self.best_ft_list

    def save(self, save_path='./model/backward_ft.dict'):
        save(save_path, self.ft_result_dict)


class FeatureSelector:
    def __init__(self, df=None, label=None, time_filed=None, val_size=0.3, nfold=5,
                 params=None, num_leaf_list=[6], metrics='logloss'):
        '''
        :param df: df, 数据集.
        :param label: str, 目标变量.
        :param time_filed: str, 时间变量.
        :param val_size: float, 验证集大小, 默认0.3.
        :param nfold: int, cv折数, 默认5.
        :param params: dict, LGB参数.
        :param num_leaf_list: list[int], 叶子数量, 在筛选的时候选择一个最好的叶子数量, 默认[4].
        '''
        self.df = df.reset_index(drop=True)
        self.label = label
        self.ft_list = df.columns.tolist()
        self.ft_list.remove(label)
        self.time_filed = time_filed
        self.val_size = val_size
        self.nfold = nfold
        self.model = None
        self.num_leaf_list = num_leaf_list

        if time_filed:
            self.ft_list.remove(self.time_filed)
            tmp_data = self.df.sort_values(by=self.time_filed)
            self.train = tmp_data.iloc[: int(len(df) * (1 - self.val_size))]
            self.val = tmp_data.iloc[int(len(df) * (1 - self.val_size)):]
            del tmp_data
        else:
            self.train, self.val = train_test_split(self.df, test_size=self.val_size,
                                                    random_state=777, stratify=self.df[self.label])

        self.df_weight = compute_sample_weight(class_weight='balanced', y=self.df[label])
        self.train_weight = compute_sample_weight(class_weight='balanced', y=self.train[label])
        self.val_weight = compute_sample_weight(class_weight='balanced', y=self.val[label])

        if metrics == 'auc':
            self.metrics = 'auc'
        else:
            self.metrics = 'binary_logloss'
        self.params = {'objective': 'binary',
                       'num_tree': 10000,
                       'learning_rate': 0.01,
                       'max_depth': -1,
                       'num_leaf': 6,
                       'bagging_freq': 1,
                       'bagging_fraction': 0.8,
                       'feature_fraction': 0.8,
                       'min_data_in_leaf': 1,
                       'min_sum_hessian_in_leaf': 0.001,
                       'lambda_l1': 0.,
                       'lambda_l2': 0.,
                       'min_split_gain': 0.,
                       # 'is_unbalance': True,
                       'seed': 777,
                       'num_thread': 4,
                       'verbose': -1,
                       'metrics': self.metrics
                       }
        if params:
            for k in params:
                self.params[k] = params[k]

        self.select_ft_list = self.ft_list
        self.select_ft_list_dict = {}

    def neg_logloss(self, y_ture, y_pred):
        weight = compute_sample_weight(class_weight='balanced', y=y_ture)
        return -log_loss(y_true=y_ture, y_pred=y_pred, sample_weight=weight)

    def hold_out(self, ft_list, num_leaf_list=None):
        '''
        Hold out.
        :param ft_list: list[str], 特征列表.
        :return: auc: float, AUC.
        '''
        best_result = -np.inf
        best_model = None
        train = lgb.Dataset(data=self.train[ft_list], label=self.train[self.label], weight=self.train_weight)
        val = lgb.Dataset(data=self.val[ft_list], label=self.val[self.label], weight=self.val_weight)
        if num_leaf_list is None:
            num_leaf_list = self.num_leaf_list
        for num_leaf in num_leaf_list:
            params = self.params.copy()
            params['num_leaf'] = num_leaf
            params['num_tree'] = 10000
            model = lgb.train(params=params, train_set=train, valid_sets=[val], valid_names=['val'],
                              early_stopping_rounds=20, verbose_eval=False)
            if self.metrics == 'auc':
                result = model.best_score['val']['auc']
            else:
                result = -model.best_score['val']['binary_logloss']
            if result > best_result:
                best_result = result
                best_model = model
        return best_result, best_model

    def cv_lgb(self, ft_list, num_leaf_list=None):
        '''
        CV.
        :param ft_list: list[str], 特征列表.
        :return: auc: float, AUC.
        '''
        best_result = -np.inf
        train = lgb.Dataset(data=self.df[ft_list], label=self.df[self.label], weight=self.df_weight)
        if num_leaf_list is None:
            num_leaf_list = self.num_leaf_list
        for num_leaf in num_leaf_list:
            params = self.params.copy()
            params['num_leaf'] = num_leaf
            params['num_tree'] = 10000
            cv_res = lgb.cv(params=params, train_set=train, nfold=self.nfold,
                            early_stopping_rounds=20, verbose_eval=False, seed=777)
            if self.metrics == 'auc':
                if cv_res['auc-mean'][-1] > best_result:
                    best_result = cv_res['auc-mean'][-1]
            else:
                if cv_res['binary_logloss-mean'][-1] > best_result:
                    best_result = cv_res['binary_logloss-mean'][-1]
        return best_result

    def cv_lr(self, ft_list):
        model = LogisticRegressionCV(Cs=10, solver='liblinear', class_weight='balanced',
                                     n_jobs=4, random_state=777, cv=5, scoring='neg_log_loss')
        model.fit(self.df[ft_list].fillna(0), self.df[self.label])
        result = model.scores_
        return np.mean(result[1], axis=0).max()

    @timer
    def boruta_select(self, perc=60, verbose=2, max_iter=100, is_ergodic=True):
        '''
        利用Boruta筛选特征.
        基模型为随机森林.
        :param perc: int, 默认90, 统计阈值, 越小得到的特征越多.
        :param verbose: int, 默认为2, 打印信息模式.
        :param max_iter: int, 默认100, 最大迭代轮数.
        :return: None.
        '''

        if not is_ergodic:
            ft_selector = BorutaPy(perc=perc, n_estimators='auto', verbose=verbose,
                                   max_iter=max_iter, random_state=777)
            train = self.df[self.select_ft_list].values

            target = self.df[self.label].values
            ft_selector.fit(train, target)
            select_ft_list = [ft for ft, support in zip(self.select_ft_list, ft_selector.support_) if support == True]
            print('-' * 20)
            print('Boruta select %d features from %d features.' % (len(select_ft_list), len(self.select_ft_list)))
            self.select_ft_list = select_ft_list
        if is_ergodic:
            for p in range(40, 90, 10):
                ft_selector = BorutaPy(perc=p, n_estimators='auto', verbose=verbose,
                                       max_iter=max_iter, random_state=777)
                train = self.df[self.ft_list].values

                target = self.df[self.label].values
                ft_selector.fit(train, target)
                self.select_ft_list_dict[p] = [ft for ft, support in zip(self.select_ft_list, ft_selector.support_) if
                                               support == True]
                print('Boruta select %d features from %d features.' % (
                    len(self.select_ft_list_dict[p]), len(self.ft_list)))

    @timer
    def null_imp_select(self, n_iter=50, num_leaf=128, perc=60, max_depth=5,
                        imp_type='gain', boosting='rf', is_plot=False, is_ergodic=True,
                        verbose=1):
        params = self.params.copy()
        params['boosting'] = boosting
        params['num_leaf'] = num_leaf
        params['max_depth'] = max_depth
        n_ft = len(self.ft_list)
        params['num_tree'] = int(100 * (n_ft / (np.sqrt(n_ft) * params['max_depth'])))
        params['bagging_fraction'] = 0.7
        params['feature_fraction'] = np.sqrt(n_ft) / n_ft

        self.df['fake_label'] = self.df[self.label].values

        # 不打乱标签列得到特征重要性.
        if verbose:
            print('正在训练模型.')
        train = lgb.Dataset(data=self.df[self.ft_list], label=self.df[self.label], weight=self.df_weight)
        model = lgb.train(params=params, train_set=train)
        ft_real_imp_list = model.feature_importance(importance_type=imp_type)
        # 打乱标签列得到特征重要性.
        np.random.seed(777)
        ft_fake_imp_list = []
        for i in range(n_iter):
            if verbose:
                print('正在训练第%d轮模型.' % (i + 1))
            np.random.shuffle(self.df['fake_label'])
            df_weight = compute_sample_weight(class_weight='balanced', y=self.df['fake_label'])
            train = lgb.Dataset(data=self.df[self.ft_list], label=self.df['fake_label'], weight=df_weight)
            model = lgb.train(params=params, train_set=train)
            ft_fake_imp_list.append(model.feature_importance(importance_type=imp_type))

        # 删除打乱的标签列.
        del self.df['fake_label']

        # 计算null_imp指标.
        ft_fake_imp_array = np.array(ft_fake_imp_list)
        null_imp_list = []
        for i in range(len(self.ft_list)):
            null_imp_list.append(np.log((1e-8 + ft_real_imp_list[i]) /
                                        (1e-8 + np.percentile(ft_fake_imp_array[:, i], perc))))
        # 根据指标筛选特征.
        self.select_ft_list = [ft for i, ft in enumerate(self.ft_list) if null_imp_list[i] > 0]
        if is_ergodic:
            for p in range(40, 90, 10):
                null_imp_list = []
                for i in range(len(self.ft_list)):
                    null_imp_list.append(np.log((1e-8 + ft_real_imp_list[i]) /
                                                (1e-8 + np.percentile(ft_fake_imp_array[:, i], p))))
                self.select_ft_list_dict[p] = [ft for i, ft in enumerate(self.ft_list) if null_imp_list[i] > 0]
                print('Null importance select %d features from %d features with perc %d.' %
                      (len(self.select_ft_list_dict[p]), len(self.ft_list), p))
        else:
            print('Null importance select %d features from %d features with perc %d.' %
                  (len(self.select_ft_list), len(self.ft_list), perc))

        # 画图.
        if is_plot:
            null_imp_list.sort(reverse=True)
            sns.barplot(x=null_imp_list, y=list(range(len(null_imp_list))), orient='h')
            plt.show()

    @timer
    def eli5_permutation_select(self, mode='cv', n_iter=50, num_leaf=128, max_depth=5,
                                boosting='rf', perc=40, is_ergodic=True):
        params = self.params.copy()
        params['boosting'] = boosting
        params['num_leaf'] = num_leaf
        params['max_depth'] = max_depth
        n_ft = len(self.ft_list)
        params['num_tree'] = int(100 * (n_ft / (np.sqrt(n_ft) * params['max_depth'])))
        params['bagging_fraction'] = 0.7
        params['feature_fraction'] = np.sqrt(n_ft) / n_ft

        if mode == 'cv':
            score_list_list = []
            skf = StratifiedKFold(n_splits=5, shuffle=True, random_state=777)
            for train_idx, val_idx in skf.split(self.df, self.df[self.label]):
                train = self.df.loc[train_idx]
                val = self.df.loc[val_idx]

                weight = compute_sample_weight(class_weight='balanced', y=train[self.label])
                train = lgb.Dataset(data=train[self.ft_list], label=train[self.label], weight=weight)
                model = lgb.train(params=params, train_set=train)

                def eli5_score_func(x, y):
                    y_pred = model.predict(x)
                    return self.neg_logloss(y, y_pred)

                _, score_list = get_score_importances(eli5_score_func, val[self.ft_list].values, val[self.label],
                                                      n_iter=n_iter, random_state=777)
                score_list = np.mean(score_list, axis=0)
                score_list_list.append(score_list)
            score_list = np.array(score_list_list).mean(axis=0)
        else:
            train = lgb.Dataset(data=self.train[self.ft_list], label=self.train[self.label], weight=self.train_weight)
            model = lgb.train(params=params, train_set=train)

            def eli5_score_func(x, y):
                y_pred = model.predict(x)
                return self.neg_logloss(y, y_pred)

            _, score_list = get_score_importances(eli5_score_func, self.val[self.ft_list].values, self.val[self.label],
                                                  n_iter=n_iter, random_state=777)
            score_list = np.mean(score_list, axis=0)
        ft_score_list = list(zip(self.ft_list, score_list))
        ft_score_list.sort(key=lambda x: x[1], reverse=True)
        ft_list = [x[0] for x in ft_score_list]
        if not is_ergodic:
            self.select_ft_list = ft_list[: int(len(ft_list) * perc / 100)]
        else:
            for p in range(30, 90, 10):
                self.select_ft_list_dict[p] = ft_list[: int(len(ft_list) * p / 100)]
        # self.select_ft_list = [ft for ft, score in zip(self.ft_list, score_list) if score > 0]

        print(len(self.select_ft_list))

    @timer
    def gain_imp_select(self, num_leaf=128, perc=60, max_depth=5,
                        boosting='rf', is_ergodic=True):
        params = self.params.copy()
        params['boosting'] = boosting
        params['num_leaf'] = num_leaf
        params['max_depth'] = max_depth
        n_ft = len(self.ft_list)
        params['num_tree'] = 100 * (n_ft / (np.sqrt(n_ft) * params['max_depth']))
        params['bagging_fraction'] = 0.7
        params['feature_fraction'] = np.sqrt(n_ft) / n_ft
        train = lgb.Dataset(data=self.df[self.ft_list], label=self.df[self.label], weight=self.df_weight)
        model = lgb.train(params=params, train_set=train)
        imp_list = model.feature_importance(importance_type='gain')
        ft_imp_list = list(zip(self.ft_list, imp_list))
        ft_imp_list.sort(key=lambda x: x[1], reverse=True)
        ft_list = [ft for ft, imp in ft_imp_list]
        if not is_ergodic:
            self.select_ft_list = ft_list[: int(len(ft_list) * perc / 100)]
        else:
            for p in range(50, 90, 10):
                self.select_ft_list_dict[p] = ft_list[: int(len(ft_list) * p / 100)]

    def same_distribution_select(self, psi_threshold=0.1, min_num_in_bin=50):
        '''
        利用训练集和验证集特征的分布是否相同进行特征筛选.
        分布是否相同利用PSI进行衡量.
        :param psi_threshold: float, 分布差异较大的阈值, 默认0.1.
        :return: None.
        '''

        def calc_psi(array_0, array_1, smooth_val=0.001):
            '''
            计算PSI.
            '''
            array_0 = array_0 * 5 + smooth_val
            array_1 = array_1 * 5 + smooth_val
            return ((array_0 - array_1) * np.log10(array_0 / array_1)).sum()

        ft_psi_dict = {}

        for ft in self.ft_list:
            train_df = self.train[[ft, self.label]].fillna(-999999).copy()
            val_df = self.val[[ft, self.label]].fillna(-999999).copy()
            # 调节训练集和验证集的正负样板比例一致.
            train_bad_rate = train_df[self.label].mean()
            val_bad_rate = val_df[self.label].mean()
            # 把逾期率小的抽取部分负样本.
            if train_bad_rate > val_bad_rate:
                bad_num = val_df[self.label].sum()
                good_num = bad_num * (1 - train_bad_rate) / train_bad_rate
                val_df_bad = val_df[val_df[self.label] == 1]
                val_df_good = val_df[val_df[self.label] == 0].sample(int(good_num), random_state=777)
                val_df = val_df_bad.append(val_df_good)
            else:
                bad_num = train_df[self.label].sum()
                good_num = bad_num * (1 - val_bad_rate) / val_bad_rate
                train_df_bad = train_df[train_df[self.label] == 1]
                train_df_good = train_df[train_df[self.label] == 0].sample(int(good_num), random_state=777)
                train_df = train_df_bad.append(train_df_good)
            # 分箱
            bins = np.quantile(train_df[ft].values, [x * 0.1 for x in range(1, 10)]).tolist()
            bins = list(sorted(set(bins)))
            bins = [-np.inf] + bins + [np.inf]
            try:
                train_df['bins'] = pd.cut(train_df[ft], bins, duplicates='drop')
                val_df['bins'] = pd.cut(val_df[ft], bins, duplicates='drop')
            except:
                ft_psi_dict[ft] = psi_threshold + 1
                continue
            train_df_g = train_df.groupby('bins').agg({self.label: ['count', 'mean']})
            train_df_g.columns = ['_'.join(x) for x in train_df_g.columns.ravel()]

            val_df_g = val_df.groupby('bins').agg({self.label: ['count', 'mean']})
            val_df_g.columns = ['_'.join(x) for x in val_df_g.columns.ravel()]

            train_df_g = train_df_g[train_df_g[self.label + '_count'] >= min_num_in_bin]
            val_df_g = val_df_g[val_df_g[self.label + '_count'] >= min_num_in_bin]

            merge_df = pd.merge(left=train_df_g, right=val_df_g, on='bins', how='outer').fillna(0)
            train_array = merge_df[self.label + '_mean_x'].values
            val_array = merge_df[self.label + '_mean_y'].values

            # 计算PSI
            ft_psi_dict[ft] = calc_psi(train_array, val_array)
        # 按阈值筛选同分布特征.
        self.select_ft_list = [x for x in ft_psi_dict if ft_psi_dict[x] <= psi_threshold]
        print('利用同分布选择%d个特征.' % len(self.select_ft_list))

    def rfe_ho(self, ft_imp_mode='gain', num_leaf_list=None):
        '''
        根据特征的重要性进行特征筛选.
        特征重要性可选模式:
            1. split
            2. gain
            3. shap
        默认gain.
        '''

        # 定序字典用于记录信息. {order: (result, ft_list), ...}
        result_dict = OrderedDict()
        ft_list = self.ft_list.copy()
        for i in range(len(ft_list)):
            result, model = self.hold_out(ft_list, num_leaf_list=num_leaf_list)
            if ft_imp_mode == 'split':
                imp_list = model.feature_importance(importance_type='split')
            elif ft_imp_mode == 'gain':
                imp_list = model.feature_importance(importance_type='gain')
            elif ft_imp_mode == 'shap':
                imp_list = None
                pass
            else:
                print('亲亲, 没有这种计算特征重要性的模式呢0.0')
                return None
            # 记录信息.
            result_dict[i] = (result, ft_list.copy())
            print('选择%d个特征, 评估结果为: %f.' % (len(ft_list), result))
            # 找到最小重要性特征.
            min_imp_ft = sorted(zip(ft_list, imp_list), key=lambda x: x[1])[0][0]
            # 移除最小重要性特征.
            ft_list.remove(min_imp_ft)
        best_result_ft = sorted(result_dict.items(), key=lambda x: x[1])[-1]

        print('最佳选择%d个特征, 评估结果为: %f.' % (len(best_result_ft[1][1]), best_result_ft[1][0]))
        self.select_ft_list = best_result_ft[1][1]

    @timer
    def lr_l1_select(self):
        '''
        利用L1正则进行特征选择.
        利用CV训练一个最优的LR模型, 提取其系数不为0的特征.
        '''
        model = LogisticRegressionCV(Cs=20,
                                     cv=5,
                                     penalty='l1',
                                     scoring='neg_log_loss',
                                     solver='liblinear',
                                     class_weight='balanced',
                                     max_iter=100,
                                     n_jobs=4,
                                     random_state=777,
                                     verbose=0)
        model.fit(X=self.df.drop(self.label, axis=1),
                  y=self.df[self.label])
        coef = model.coef_[0]
        self.select_ft_list = [x for i, x in enumerate(self.ft_list) if coef[i] != 0]
        print('L1逻辑回归在%d个特征中选择%d个特征.' % (len(self.ft_list), len(self.select_ft_list)))

    def set_metrics(self, metrics):
        self.metrics = metrics
        self.params['metrics'] = metrics

    def set_params(self, params):
        for p in params:
            if p in self.params:
                self.params[p] = params[p]

    def save_feature(self, path='select_ft', is_save_ft_list=True,
                     is_save_ft_list_dict=False):
        '''
        保存特征.
        :param path: str, 特征保存路径.
        :return: None
        '''
        if is_save_ft_list_dict:
            with open(path + '_list.pkl', 'wb') as f:
                pickle.dump(self.select_ft_list_dict, f)
                f.close()
        if is_save_ft_list:
            with open(path + '.pkl', 'wb') as f:
                pickle.dump(self.select_ft_list, f)
                f.close()


class TraversalSearchSelector:
    def __init__(self, df=None,
                 label='target',
                 nfold=7,
                 ft_num_range=(7, 8),
                 verbose=10):
        self.df = df
        self.label = label
        self.ft_list = df.columns.tolist()
        if label in self.ft_list:
            self.ft_list.remove(label)
        else:
            print('警告, 没有目标列.')

        self.ft_num_range = ft_num_range
        self.cv = LRCV(nfold=nfold)
        self.result_dict = {}
        self.best_ft_list = None
        self.best_result = -np.inf
        self.verbose = verbose


    @timer
    def run(self):
        '''遍历搜索特征'''
        # 数据标准化.
        all_ft_list = self.ft_list.copy()
        for i, ft_list in enumerate(traversal_search(all_ft_list, n_range=self.ft_num_range)):
            cv_result = self.cv.cv(X=self.df[ft_list], y=self.df[self.label])
            self.result_dict[i] = (len(ft_list), cv_result, ft_list)
            self.best_result = cv_result if cv_result > self.best_result else self.best_result
            self.best_ft_list = ft_list if cv_result > self.best_result else self.best_ft_list
            if i % self.verbose == 0:
                print('第%d轮, 特征数量%d, CV结果为%f, 最佳CV结果为%f' % (i, len(ft_list), cv_result, self.best_result))

    def save(self, path='./model/traversal_search_ft.dict'):
        save(path=path, obj=self.result_dict)


class TraversalSearchSelectorWithIndicator:
    def __init__(self, df=None,
                 label='target',
                 nfold=5,
                 ft_num_range=(5, 6, 7),
                 verbose=10):
        self.df = df
        self.label = label
        self.ft_list = df.columns.tolist()
        if label in self.ft_list:
            self.ft_list.remove(label)
        else:
            print('警告, 没有目标列.')
        self.indicate_ft_list = []
        for ft in self.ft_list:
            if ft.endswith('_indicate'):
                self.indicate_ft_list.append(ft)

        for ft in self.indicate_ft_list:
            self.ft_list.remove(ft)
        print('包含的特征有%d个' % len(self.ft_list))
        print('包含指示列的特征有%d个.' % len(self.indicate_ft_list))

        self.ft_num_range = ft_num_range
        self.cv = LRCV(nfold=nfold)
        self.result_dict = {}
        self.best_ft_list = None
        self.best_result = -np.inf
        self.verbose = verbose


    @timer
    def run(self):
        '''遍历搜索特征'''
        # 数据标准化.
        all_ft_list = self.ft_list.copy()
        for i, ft_list in enumerate(traversal_search(all_ft_list, n_range=self.ft_num_range)):
            n_ft = len(ft_list)
            indicate_list = []
            for ft in ft_list:
                if ft + '_indicate' in self.indicate_ft_list:
                    indicate_list.append(ft + '_indicate')
            ft_list = ft_list + indicate_list
            cv_result = self.cv.cv(X=self.df[ft_list], y=self.df[self.label])
            self.result_dict[i] = (n_ft, cv_result, ft_list)
            self.best_result = cv_result if cv_result > self.best_result else self.best_result
            self.best_ft_list = ft_list if cv_result > self.best_result else self.best_ft_list
            if i % self.verbose == 0:
                print('第%d轮, 特征数量%d, CV结果为%f, 最佳CV结果为%f' % (i, n_ft, cv_result, self.best_result))

    def save(self, path='./model/traversal_search_indicate_ft.dict'):
        save(path=path, obj=self.result_dict)



if __name__ == '__main__':
    train = pd.read_csv('./data/bc_train.csv')
    is_ = ModelParamsGenerator(df=train, label='target', nfolds=3,
                               ft_percs=(100, 60, 30), init_points=5, n_iter=0)
    is_.run()



# coding=utf-8
'''
- 数据预处理类.
- 其中大部分处理方式是已经被封装好的, 直接使用.
- 若要反复使用, 利用save(path='../transformer/...', obj=...)保存.
- 少部分处理方式自己根据需求实现.
- 可用转换器:
    1. Normalizer
        nmtf = Normalizer(norm='l2', copy=True)

    2. StandardScaler
        stdtf = StandardScaler(copy=True, with_mean=True, with_std=True)

    3. RobustScaler
        rbtf = RobustScaler(with_centering=True, with_scaling=True, quantile_range=(25.0, 75.0), copy=True)

    4. PowerTransformer
        ptf = PowerTransformer(method=’yeo-johnson’, standardize=True, copy=True)

    5. QuantileTransformer
        qttf = QuantileTransformer(n_quantiles=1000, output_distribution=’uniform’, random_state=None, copy=True)

    6. CatMeanEncoder
        cmecd = CatMeanEncoder(cat_ft_list=None, label=None)

    7. MissingValueImputer
        msvi = MissingValueImputer(missing_values=np.nan, strategy='constant',
            fill_value=None, verbose=0, model_type='gbdt', miss_perc_range=(0, 1))

'''

import numpy as np
import pandas as pd
from scipy.stats import binom
from collections import OrderedDict

from sklearn.experimental import enable_iterative_imputer
from sklearn.impute import SimpleImputer
from sklearn.impute._iterative import IterativeImputer
from sklearn.ensemble import RandomForestRegressor
from sklearn.model_selection import StratifiedKFold
from sklearn.preprocessing import StandardScaler, PowerTransformer, MinMaxScaler

from .base_utils import save

import warnings

warnings.filterwarnings('ignore')


class MissingValueImputerWithTarget:
    '''
    缺失值填充, 主要用于生产环境融合模型.
    '''
    def __init__(self, min_miss_num=100, is_indicate=True, mode='mean'):
        self.is_indicate = is_indicate
        self.ft_list = None
        self.min_miss_num = min_miss_num
        if mode not in ('left', 'right', 'mean'):
            print('模式设定不正确, 目前只支持: left, right, mean')
        self.mode = mode
        self.map_dict = OrderedDict()
        self.miss_ft_set = set()

    def fit(self, X=None, y=None):
        self.ft_list = X.columns.tolist()
        for ft in self.ft_list:
            # 若缺失数量少, 则补中位数.
            miss_num = X[ft].isnull().sum()
            if miss_num > 0:
                self.miss_ft_set.add(ft)
            if miss_num < self.min_miss_num:
                self.map_dict[ft] = X[ft].median()
                continue
            # 若缺失样本多, 则根据分箱后的信息进行填充.
            df_copy = X[[ft]].copy()
            df_copy['target'] = y
            df_copy_miss = df_copy[df_copy[ft].isnull()]
            df_copy_notna = df_copy[df_copy[ft].notna()]
            # 计算缺失部分的逾期率.
            miss_overdue_rate = df_copy_miss['target'].mean()
            # 对非缺失进行分箱, 得到每箱的缺失率.
            df_copy_notna['bin'] = pd.qcut(df_copy_notna['target'], 10, precision=8, duplicates='drop')
            df_copy_notna_g = df_copy_notna.groupby('bin').agg(overdue_mean=('target', 'mean'))
            df_copy_notna_g.reset_index(inplace=True)
            # 找到与缺失值数据最接近的分箱, 根据预设模式选择填充值.
            df_copy_notna_g['abs_sub'] = df_copy_notna_g['overdue_mean'].apply(lambda x: abs(x - miss_overdue_rate))
            min_abs_sub = min(df_copy_notna_g['abs_sub'].values)
            interval = df_copy_notna_g.loc[df_copy_notna_g['abs_sub'] == min_abs_sub, 'bin'].values[0]
            if self.mode == 'left':
                self.map_dict[ft] = interval.left
            elif self.mode == 'right':
                self.map_dict[ft] = interval.right
            else:
                self.map_dict[ft] = (interval.right + interval.left) / 2

        return self

    def transform(self, X=None):
        for ft in self.map_dict:
            if ft in self.miss_ft_set and self.is_indicate:
                X[ft + '_indicate'] = X[ft].apply(lambda x: int(pd.isnull(x)))
            X[ft] = X[ft].fillna(self.map_dict[ft])
        return X

    def fit_transform(self, X=None, y=None):
        self.fit(X=X, y=y)
        return self.transform(X=X)

    def save(self, name='miss.imputer'):
        save(path='./transformer/' + name, obj=self)
        print('成功保存.')


class StandardTransformer:
    def __init__(self):
        self.std_scl = StandardScaler()
        self.ft_list = None

    def fit(self, X=None, y=None):
        self.ft_list = X.columns.tolist()
        self.std_scl.fit(X=X)
        return self

    def transform(self, X=None):
        X_copy = self.std_scl.transform(X=X[self.ft_list])
        X_copy = pd.DataFrame(X_copy, columns=self.ft_list)
        return X_copy

    def fit_transform(self, X=None, y=None):
        self.fit(X=X, y=y)
        return self.transform(X=X)

    def save(self, name='std.transformer'):
        save(path='./transformer/' + name, obj=self)
        print('成功保存.')


class MissingValueImputer:
    '''
    对scikit-learn的缺失值填充器的封装.
    根据自己对缺失值填充的常用需求以及理解进行改进.
    '''

    def __init__(self,missing_values=np.nan, strategy='constant',
                 fill_value=None, verbose=0, initial_strategy='median'):
        self.missing_values = missing_values
        self.strategy = strategy
        self.fill_value = fill_value
        self.initial_strategy = initial_strategy
        if strategy not in ('constant', 'mean', 'median', 'predict'):
            print('亲亲您的填充方法没有输入对呢, 这边只支持如下方法嗷~\n'
                  '1. constant\n'
                  '2. mean\n'
                  '3. median\n'
                  '4. predict')
        # 初始化填充器.
        if self.strategy != 'predict':
            self.tf = SimpleImputer(strategy=self.strategy, fill_value=self.fill_value, copy=True,
                                    add_indicator=True, verbose=verbose)
        elif self.strategy == 'predict':
            self.model = RandomForestRegressor(random_state=777, max_depth=4, max_features='sqrt')
            self.tf = IterativeImputer(estimator=self.model, initial_strategy=initial_strategy,
                                       imputation_order='ascending',
                                       random_state=777, add_indicator=True, verbose=verbose)
        self.ft_list = None
        self.ft_miss_list = None

    def fit(self, X=None, y=None):
        X.replace(to_replace=self.missing_values, value=np.nan, inplace=True)
        self.ft_list = X.columns.tolist()
        ft_miss_perc = X.isnull().mean(axis=0)
        self.ft_miss_list = ft_miss_perc[ft_miss_perc > 0].index.tolist()
        if self.strategy == 'predict':
            n_ft = len(self.ft_list)
            self.model.set_params(n_estimators=int(100 * (n_ft / (np.sqrt(n_ft) * self.model.max_depth))))
            self.tf.set_params(estimator=self.model)
        self.tf.fit(X=X)
        print('填充缺失值训练完毕.')
        return self

    def transform(self, X=None):
        X.replace(to_replace=self.missing_values, value=np.nan, inplace=True)
        copy_df = self.tf.transform(X[self.ft_list])
        indicat_col = [col + '_indicat' for col in self.ft_miss_list]
        copy_df = pd.DataFrame(copy_df, columns=self.ft_list + indicat_col)
        for col in copy_df:
            X[col] = copy_df[col]
        return X

    def fit_transform(self, X=None, y=None):
        self.fit(X=X, y=y)
        print('缺失值训练填充完毕.')
        return self.transform(X=X)

    def save(self, name='miss_val.imputer'):
        save(path='./transformer/' + name, obj=self)
        print('成功保存.')


class BoxCoxTransformer:
    def __init__(self, boxcox_mode='yeo-johnson'):
        self.minmax = MinMaxScaler(feature_range=(2, 3))
        self.boxcox = PowerTransformer(method=boxcox_mode, standardize=False)
        self.ft_list = None

    def fit(self, X=None, y=None):
        # 记录特征名称.
        self.ft_list = X.columns.tolist()

        # 进行minmax.
        X = self.minmax.fit_transform(X)
        X = pd.DataFrame(X, columns=self.ft_list)

        # boxcox拟合.
        self.boxcox.fit(X, y)
        return self

    def transform(self, X=None):
        # 预处理.
        X = X[self.ft_list]
        X = self.minmax.transform(X)
        X = pd.DataFrame(X, columns=self.ft_list)
        # boxcox变换.
        X = self.boxcox.transform(X)
        X = pd.DataFrame(X, columns=self.ft_list)
        return X

    def fit_transform(self, X=None, y=None):
        self.fit(X=X, y=y)
        return self.transform(X=X)

    def save(self, name='boxcox.transformer'):
        save(path='./transformer/' + name, obj=self)
        print('boxcox.transformer成功保存.')


class LinearModelPreprocessor:
    def __init__(self,
                 is_boxcox=True,
                 boxcox_mode='yeo-johnson',
                 is_std=True,
                 is_miss_impute=True,
                 min_miss_num=100,
                 miss_mode='mean',
                 is_indicate=True):
        self.is_boxcox = is_boxcox
        self.is_std = is_std
        self.is_miss_impute = is_miss_impute

        if self.is_boxcox:
            self.boxcox = BoxCoxTransformer(boxcox_mode=boxcox_mode)
        if self.is_std:
            self.std = StandardTransformer()
        if is_miss_impute:
            self.miss_impute = MissingValueImputerWithTarget(min_miss_num=min_miss_num,
                                                             mode=miss_mode,
                                                             is_indicate=is_indicate)
        self.ft_list = None

    def fit(self, X=None, y=None):
        self.ft_list = X.columns.tolist()
        if self.is_boxcox:
            X = self.boxcox.fit_transform(X)
        if self.is_std:
            X = self.std.fit_transform(X)
        if self.is_miss_impute:
            self.miss_impute.fit(X, y)
        return self

    def transform(self, X=None):
        X = X[self.ft_list]
        if self.is_boxcox:
            X = self.boxcox.transform(X)
        if self.is_std:
            X = self.std.transform(X)
        if self.is_miss_impute:
            X = self.miss_impute.transform(X)
        return X

    def fit_transform(self, X=None, y=None):
        self.fit(X=X, y=y)
        return self.transform(X=X)

    def save(self, name='lilear.preprocessor'):
        save(path='./transformer/' + name, obj=self)
        print('成功保存.')


class CategoryMeanEncoder:
    def __init__(self, cat_ft_list=None, alpha_perc=0.1, n_splits=5):
        self.alpha_perc = alpha_perc
        self.cat_ft_list = cat_ft_list
        self.n_splits = n_splits
        self.skf = StratifiedKFold(n_splits=n_splits, shuffle=True, random_state=777)
        self.ft_attr_encode_dict = {}
        self.global_mean = None
        self.alpha = None

    def fit(self, X=None, y=None):
        self.ft_attr_encode_dict = {}
        self.global_mean = y.mean()
        self.alpha = self.alpha_perc * X.shape[0] * (1 - 1 / self.n_splits)
        cat_ft_list = X.select_dtypes('object').columns.tolist()
        if self.cat_ft_list is None:
            self.cat_ft_list = cat_ft_list
        else:
            for ft in cat_ft_list:
                if ft not in self.cat_ft_list:
                    self.cat_ft_list.append(ft)

        for ft in self.cat_ft_list:
            tmp_list = []
            for train_idx, _ in self.skf.split(X, y):
                train = X.loc[train_idx]
                tmp_df = train[[ft]].copy()
                tmp_df['target'] = y
                tmp_df_g = tmp_df.groupby(ft).agg(mean=('target', 'mean'),
                                                  count=('target', 'count'))

                tmp_df_g['mean_smooth'] = tmp_df_g.apply(lambda x:
                                                         (x['mean'] * x['count'] + self.global_mean * self.alpha) /
                                                         (x['count'] + self.alpha), axis=1)
                tmp_list.append(tmp_df_g[['mean_smooth']])
            attr_mean_df = pd.concat(tmp_list, axis=1, join='outer')
            attr_mean_df = attr_mean_df.fillna(self.global_mean)
            attr_mean_df['result'] = attr_mean_df.values.sum(axis=1)
            self.ft_attr_encode_dict[ft] = dict(zip(attr_mean_df.index.values, attr_mean_df['result'].values))
        print('均值编码训练完毕.')
        return self

    def transform(self, X=None):
        for ft in self.cat_ft_list:
            X[ft] = X[ft].map(self.ft_attr_encode_dict[ft])
        print('均值编码转换完毕.')
        return X

    def fit_transform(self, X=None, y=None):
        self.fit(X=X, y=y)
        return self.transform(X=X)


    def save(self, name='bys_cat_mean_encoder.pkl'):
        save(path='../transformer/' + name, obj=self)
        print('成功保存.')


class CategoryPreprocessor:
    '''
    主要针对类别特征高基数, 或者存在某属性数量极少的情况, 进行预处理.
    不管缺失值, 等到缺失值填充步骤处理.
    利用二项分布找出显著的属性, 先保持不动.
    设定最小占比阈值, 尝试将小于该阈值的且不显著属性随机进行合并.
    若合并后还剩余小于阈值的, 则将其与另外不显著的最大属性合并.
    若没有, 和显著的最大属性合并.
    检查是否单值, 是则删除该特征.
    '''
    def __init__(self,
                 min_perc=0.05,
                 p_value=0.0001,
                 cat_ft_list=None,
                 label='target'):

        self.min_perc = min_perc
        self.p_value = p_value
        self.cat_ft_list = cat_ft_list
        self.label = label
        self.ft_attr_map_dict = {}


    def fit(self, X=None):
        np.random.seed(777)
        # 收集需要预处理的类别特征.
        if self.cat_ft_list is None:
            self.cat_ft_list = X.select_dtypes('object').columns.tolist()
        else:
            for col in X.select_dtypes('object').columns.tolist():
                if col not in self.cat_ft_list:
                    self.cat_ft_list.append(col)
        # 根据二项分布过滤出显著属性.
        len_X = X.shape[0]
        min_num_in_attr = len_X * self.min_perc
        global_mean = X[self.label].mean()

        def is_significant(data, count):
            if data > binom.ppf(1 - self.p_value, count, global_mean):
                return True
            elif data < binom.ppf(self.p_value, count, global_mean):
                return True
            else:
                return False
        def comb(attr_list, info_df, min_num):
            np.random.shuffle(attr_list)
            bin_list = []
            # 遍历.
            for i in range(len(attr_list)):
                attr = attr_list.pop()
                if len(bin_list) == 0 or bin_list[-1][1] > min_num:
                    bin_list.append([[attr], info_df.loc[attr, 'count_']])
                else:
                    bin_list[-1][0].append(attr)
                    bin_list[-1][1] += info_df.loc[attr, 'count_']
            # 检查.
            if len(bin_list) == 1 and bin_list[0][1] < min_num_in_attr:
                return bin_list, False
            elif len(bin_list) == 1 and bin_list[0][1] >= min_num_in_attr:
                return bin_list, True
            elif bin_list[-1][1] < min_num:
                attr_info = bin_list.pop()
                bin_list[-1][0] = bin_list[-1][0] + attr_info[0]
                bin_list[-1][1] += attr_info[1]
                return bin_list, True
            else:
                return bin_list, True

        for ft in self.cat_ft_list:
            tmp_df = X[[ft, self.label]].copy()
            tmp_df_g = tmp_df.groupby(ft).agg(count_=(self.label, 'count'),
                                              sum_=(self.label, 'sum'))
            tmp_df_g['is_significant'] = tmp_df_g.apply(lambda x: is_significant(x['sum_'], x['count_']), axis=1)
            main_attr = [x for x, y in zip(tmp_df_g.index, tmp_df_g['is_significant']) if y is True]
            not_main_attr = [x for x, y in zip(tmp_df_g.index, tmp_df_g['is_significant']) if y is False]
            # 对不显著, 且数量小于阈值的属性进行合并.
            not_main_small_attr = [x for x in not_main_attr if tmp_df_g.loc[x, 'count_'] < min_num_in_attr]
            not_main_big_attr = [x for x in not_main_attr if tmp_df_g.loc[x, 'count_'] >= min_num_in_attr]
            bin_list = None
            if not_main_small_attr:
                bin_list, is_good = comb(not_main_small_attr, tmp_df_g, min_num_in_attr)
                # 与不显著中最大的属性合并.
                if not is_good:
                    bigest_attr = None
                    bigest_num = 0
                    if len(not_main_big_attr) != 0:
                        for attr in not_main_big_attr:
                            if tmp_df_g.loc[attr, 'count_'] > bigest_num:
                                bigest_num = tmp_df_g.loc[attr, 'count_']
                                bigest_attr = attr
                        not_main_big_attr.remove(bigest_attr)
                        bin_list[-1][0].append(bigest_attr)
                    # 与显著属性中最大的合并.
                    else:
                        for attr in main_attr:
                            if tmp_df_g.loc[attr, 'count_'] > bigest_num:
                                bigest_num = tmp_df_g.loc[attr, 'count_']
                                bigest_attr = attr
                        main_attr.remove(bigest_attr)
                        bin_list[-1][0].append(bigest_attr)
            # 构建映射字典.
            self.ft_attr_map_dict[ft] = {}
            for attr in main_attr:
                self.ft_attr_map_dict[ft][attr] = attr
            for attr in not_main_big_attr:
                self.ft_attr_map_dict[ft][attr] = attr
            if bin_list is not None:
                for attr_list, _ in bin_list:
                    for attr in attr_list:
                        attr_name = ''.join(attr_list)
                        self.ft_attr_map_dict[ft][attr] = attr_name
        return self

    def transform(self, X=None):
        # 根据字典进行映射.
        for ft in self.ft_attr_map_dict:
            X[ft] = X[ft].map(self.ft_attr_map_dict[ft])
            # 转换为0-1特征.
            if X[ft].nunique(dropna=False) == 2:
                unique_tuple = X[ft].unique()
                map_dict = {unique_tuple[0]: 0, unique_tuple[1]: 1}
                X[ft] = X[ft].map(map_dict)
            if X[ft].nunique(dropna=False) == 1:
                del X[ft]
        return X

    def fit_transform(self, X=None):
        self.fit(X=X)
        return self.transform(X=X)



if __name__ == '__main__':

    from param_optimize import LightBayesOptimizer
    from sklearn.metrics import roc_auc_score
    train = pd.read_csv('./data/hr_train.csv')
    test = pd.read_csv('./data/hr_test.csv')
    me = CategoryMeanEncoder(label='target')
    train = me.fit_transform(train)
    test = me.transform(test)

    po = LightBayesOptimizer(df=train, label='target')
    po.run()
    po.fit_model()

    print(roc_auc_score(train['target'], po.predict(train)))
    print(roc_auc_score(test['target'], po.predict(test)))


    pass
'''
bc miss
lr
0
0.9955830753353974
0.9914021164021163
mean
0.9962641898864809
0.9914021164021165
median
0.9961197110423117
0.9904100529100529

lgb
miss
0.9997729618163054
0.9976851851851852
0 lr
0.999938080495356
0.9976851851851851
mean lr
0.9992569659442725
0.9993386243386243
median lr
0.9999793601651187
0.9986772486772486
predict mean gbdt
0.9998348813209493
0.9993386243386243
predict mean lr
0.9994840041279669
0.9996693121693121
'''



# coding=utf-8
'''
GBDT, LR.
'''
import numpy as np
import pandas as pd

from sklearn.metrics import roc_auc_score
from sklearn.model_selection import train_test_split
from sklearn.utils.class_weight import compute_sample_weight
from sklearn.linear_model import LogisticRegression

import lightgbm as lgb


from .base_utils import save, timer, neg_log_loss, split_X_y, load
from .data_transform import StandardTransformer
from .data_visualization import Echart

import os
import warnings

warnings.filterwarnings('ignore')


class LR:
    def __init__(self, df=None, label='target',
                 penalty='l2', C=1, is_std=True):
        self.label = label
        self.ft_list = df.columns.tolist()
        if label in self.ft_list:
            self.ft_list.remove(label)
        self.model = LogisticRegression(penalty=penalty,
                                        C=C,
                                        class_weight='balanced',
                                        random_state=7,
                                        solver='liblinear',
                                        n_jobs=4)

        self.X, self.y = split_X_y(df=df, label=label)
        if is_std:
            std_scl = StandardTransformer()
            self.X = std_scl.fit_transform(X=self.X)


    def run(self):
        self.model.fit(X=self.X, y=self.y)


    def predict(self, X=None):
        return self.model.predict_proba(X=X[self.ft_list])[:, 1]


    def score(self, X=None, y=None, metric='auc'):
        if metric == 'auc':
            res = roc_auc_score(y_true=y, y_score=self.predict(X=X))
            return res
        elif metric == 'logloss':
            res = neg_log_loss(y_true=y, y_pred=self.predict(X=X))
        else:
            print('警告, 评估参数输入错误.')
        return res


    def save(self, save_path='LR.model'):
        save(path='./model/' + save_path, obj=self.model)



class Light:
    def __init__(self):

        pass


    def run(self):

        pass

    def save(self):

        pass


if __name__ == '__main__':

    pass


# coding=utf-8

import numpy as np
import pandas as pd

from sklearn.model_selection import StratifiedKFold
from sklearn.linear_model import LogisticRegressionCV
from sklearn.metrics import roc_auc_score, log_loss
from sklearn.utils.class_weight import compute_sample_weight
from sklearn.preprocessing import StandardScaler

from .base_utils import save

from category_encoders.james_stein import JamesSteinEncoder
from category_encoders.target_encoder import TargetEncoder
from category_encoders.woe import WOEEncoder

from collections import OrderedDict
import warnings

np.random.seed(777)
warnings.filterwarnings('ignore')


class Stacker:
    '''
    模型融合.
    准备好sub_model_dict非常重要.
    '''

    def __init__(self, train_df=None, label=None, sub_model_dict=None, test_df=None,
                 show_sub_model=False, with_origin_ft=True, origin_ft_name=None):
        '''
        :param train_df: df, 数据集, 包含所有特征和目标列.
        :param label: str, 目标列名称.
        :param sub_model_dict: dict, 子模型信息字典. {model_name: (ft_name, model_class, params)}
        '''
        self.origin_ft_name = origin_ft_name
        self.with_origin_ft = with_origin_ft
        self.show_sub_model = show_sub_model
        self.train_df = train_df.reset_index(drop=True)
        self.label = label
        self.sub_model_dict = sub_model_dict
        self.test_df = test_df.reset_index(drop=True)

    def make_up_data(self, data=None, mode='h'):
        '''
        拼接数据.
        :param data: list[array_0, ...].
        :param mode: str, 拼接模式, 水平'h', 竖直'v'.
        :return: array(2d)
        '''
        if mode == 'h':
            for i, array in enumerate(data):
                data[i] = data[i].reshape((-1, 1))
            return np.concatenate(data, axis=1)
        else:
            return np.concatenate(data)

    def avg_array(self, data):
        '''
        计算array的平均.
        :param data: list[array_0, ...].
        :return: array.
        '''
        res_array = np.zeros_like(data[0])
        for array in data:
            res_array = res_array + array
        return res_array / len(data)

    def logloss(self, model, x, y):
        weight = compute_sample_weight(class_weight='balanced', y=y)
        return -log_loss(y, model.predict_proba(x)[:, 1], sample_weight=weight)

    def sub_model_result(self):
        for model_name in self.sub_model_dict:
            ft_name = self.sub_model_dict[model_name][0]
            model_class = self.sub_model_dict[model_name][1]
            params = self.sub_model_dict[model_name][2]
            # 训练模型.
            model = model_class(**params)
            model.run(self.train_df[ft_name], self.train_df[self.label])
            # 预测结果.
            print('%s的结果为: %f' % (
                model_name, roc_auc_score(self.test_df[self.label], model.predict_proba(self.test_df[ft_name])[:, 1])))

    def stack_two_layer(self, kfold=5):
        if self.show_sub_model:
            self.sub_model_result()
        pred_y_on_cv_list = []
        pred_y_on_test_list = []
        skf = StratifiedKFold(n_splits=kfold, random_state=777, shuffle=True)
        for model_name in self.sub_model_dict:
            print('正在训练%s模型.' % model_name)
            ft_name = self.sub_model_dict[model_name][0]
            model_class = self.sub_model_dict[model_name][1]
            params = self.sub_model_dict[model_name][2]
            # 切分数据集.
            tmp_pred_y_on_cv_list = []
            tmp_pred_y_on_test_list = []
            for train_index, test_index in skf.split(self.train_df[ft_name], self.train_df[self.label]):
                train_x, train_y = self.train_df.loc[train_index, ft_name], self.train_df.loc[train_index, self.label]
                test_x, test_y = self.train_df.loc[test_index, ft_name], self.train_df.loc[test_index, self.label]
                # 训练模型.
                model = model_class(**params)
                model.run(train_x, train_y)
                # 预测结果.
                tmp_pred_y_on_cv_list.append(model.predict_proba(test_x)[:, 1])
                tmp_pred_y_on_test_list.append(model.predict_proba(self.test_df[ft_name])[:, 1])

            # 拼接预测结果.

            pred_y_on_cv_list.append(self.make_up_data(data=tmp_pred_y_on_cv_list, mode='v'))
            pred_y_on_test_list.append(self.avg_array(tmp_pred_y_on_test_list))

        # 拼接结果.
        pred_y_on_cv_mat = self.make_up_data(data=pred_y_on_cv_list, mode='h')
        pred_y_on_test_mat = self.make_up_data(data=pred_y_on_test_list, mode='h')
        # 每个模型输出的结果.
        for i in range(pred_y_on_test_mat.shape[1]):
            print(roc_auc_score(self.test_df[self.label], pred_y_on_test_mat[:, i]))

        # 二阶模型.
        print('正在训练二阶逻辑回归模型.')
        if self.with_origin_ft:
            pred_y_on_cv_mat = np.concatenate([self.train_df[self.origin_ft_name].values, pred_y_on_cv_mat], axis=1)
            pred_y_on_test_mat = np.concatenate([self.test_df[self.origin_ft_name].values, pred_y_on_test_mat], axis=1)
        # 特征归一化.
        scl = StandardScaler()
        scl.fit(pred_y_on_cv_mat)
        pred_y_on_cv_mat = scl.transform(pred_y_on_cv_mat)
        pred_y_on_test_mat = scl.transform(pred_y_on_test_mat)

        lr_model_l1 = LogisticRegressionCV(Cs=100, cv=5, n_jobs=4,
                                           solver='liblinear', scoring=self.logloss,
                                           penalty='l1', class_weight='balanced', random_state=777)
        lr_model_l1.fit(X=pred_y_on_cv_mat, y=self.train_df[self.label])

        lr_model_l2 = LogisticRegressionCV(Cs=100, cv=5, n_jobs=4,
                                           solver='liblinear', scoring=self.logloss,
                                           penalty='l2', class_weight='balanced', random_state=777)
        lr_model_l2.fit(X=pred_y_on_cv_mat, y=self.train_df[self.label])

        # 测试集结果.
        lr_model_l1_score = roc_auc_score(self.test_df[self.label], lr_model_l1.predict_proba(pred_y_on_test_mat)[:, 1])
        lr_model_l2_score = roc_auc_score(self.test_df[self.label], lr_model_l2.predict_proba(pred_y_on_test_mat)[:, 1])
        if lr_model_l1_score > lr_model_l2_score:
            self.lr_model = lr_model_l1
            print('测试集结果: %f' % lr_model_l1_score)
        else:
            self.lr_model = lr_model_l2
            print('测试集结果: %f' % lr_model_l2_score)
        print(roc_auc_score(self.test_df[self.label], pred_y_on_test_mat.sum(axis=1)))

    def stack_three_layer(self):
        pass


class CategoryEncodeCombiner:
    '''
    对几种不同的类别特征编码方式的融合.
    遇事不决上融合.
    考虑到不同编码方式的输出数值范围不一样, 先进行标准化, 再加权融合.
    用于融合的编码方式和默认权重:
                编码方式                    权重
        1. JamesSteinEncoder - binary       50
        2. JamesSteinEncoder - beta         20
        3. TargetEncoder                    20
        4. WOEEncoder                       10
    '''

    def __init__(self, weight=None):
        self.weight = (50, 10, 20, 20) if weight is None else weight
        self.encoder_dict = OrderedDict({'JamesSteinEncoder_binary': JamesSteinEncoder(handle_missing='return_nan',
                                                                                       model='binary', random_state=7,
                                                                                       return_df=False),
                                         'JamesSteinEncoder_beta': JamesSteinEncoder(handle_missing='return_nan',
                                                                                     model='beta', random_state=7,
                                                                                     return_df=False),
                                         'TargetEncoder': TargetEncoder(handle_missing='return_nan', return_df=False),
                                         'WOEEncoder': WOEEncoder(handle_missing='return_nan', random_state=7,
                                                                  return_df=False)})
        self.stdscl_dict = OrderedDict({'JamesSteinEncoder_binary': StandardScaler(),
                                        'JamesSteinEncoder_beta': StandardScaler(),
                                        'TargetEncoder': StandardScaler(),
                                        'WOEEncoder': StandardScaler()})
        self.ft_cat_list = None

    def fit(self, X=None, y=None):
        self.ft_cat_list = X.select_dtypes('object').columns.tolist()
        for ecd in self.encoder_dict:
            X_ecd = self.encoder_dict[ecd].fit_transform(X=X[self.ft_cat_list], y=y)
            self.stdscl_dict[ecd].fit(X=X_ecd)
        return self

    def transform(self, X=None):
        X_res = 0
        for i, ecd in enumerate(self.encoder_dict):
            X_ecd = self.encoder_dict[ecd].transform(X=X[self.ft_cat_list])
            X_ecd_stdscl = self.stdscl_dict[ecd].transform(X=X_ecd)
            X_res = X_res + X_ecd_stdscl * self.weight[i]
        X_res = X_res / 100
        X_res = pd.DataFrame(X_res, columns=self.ft_cat_list)
        for col in X_res.columns.tolist():
            X[col] = X_res[col]
        return X

    def fit_transform(self, X=None, y=None):
        self.fit(X=X, y=y)
        return self.transform(X=X)

    def save(self, path='TransformCombiner.pkl'):
        save(path=path, obj=self)


if __name__ == '__main__':
    train = pd.read_csv('./data/hr_train.csv')
    train_X, train_y = train.drop('target', axis=1), train['target']
    cat_ecd_cb = CategoryEncodeCombiner()
    print(train_X.head(20))
    train_X = cat_ecd_cb.fit_transform(train_X, train_y)
    print(train_X.info())
    print(train_X.head(20))




# coding=utf-8
'''
数据可视化.
'''

import numpy as np
import pandas as pd
import pandas_profiling
import matplotlib.pyplot as plt
from matplotlib.font_manager import _rebuild
import seaborn as sns

_rebuild()
plt.rcParams['font.sans-serif'] = 'SimHei'
plt.rcParams['axes.unicode_minus'] = False

from sklearn.metrics import roc_auc_score

import pyecharts.options as opts
from pyecharts.render import make_snapshot
from pyecharts.charts import Line, Bar, Pie
from snapshot_phantomjs import snapshot
from datetime import datetime, timedelta
import re
import os
import warnings

warnings.filterwarnings('ignore')

from .base_utils import mkdir


class TargetPlot:
    def __init__(self):
        pass

    def run(self):
        pass


class LiftChart:
    def __init__(self):
        pass

    def run(self):
        pass


class PDP:
    def __init__(self):
        pass

    def run(self):
        pass


class PSI:
    def __init__(self):
        pass

    def run(self):
        pass


class ValueOnTime:
    def __init__(self):
        pass

    def run(self):
        pass


class EvaluationOnTime:
    def __init__(self, df=None, label='target', time_field='time',
                 n_bin=5,
                 time_range_list=None,
                 eval_metric='auc', save_path='./image/EvaluationOnTime/'):
        self.df = df
        self.label = label
        self.time_field = time_field
        self.n_bin = n_bin
        self.time_range_list = time_range_list
        self.eval_metric = eval_metric
        self.save_path = save_path
        mkdir(self.save_path)

        self.df = df.sort_values(self.time_field, ascending=True)
        self.ft_list = self.df.columns.tolist()
        if self.label in self.ft_list:
            self.ft_list.remove(self.label)
        else:
            print('警告, 数据不包含目标列.')
        if self.time_field in self.ft_list:
            self.ft_list.remove(self.time_field)
        else:
            print('警告, 数据不包含时间列.')
        if self.time_range_list is None:
            fst_date = self.df[self.time_field].tolist()[0]
            lst_date = self.df[self.time_field].tolist()[-1]
            fst_date = datetime.strptime(fst_date, '%Y-%m-%d')
            lst_date = datetime.strptime(lst_date, '%Y-%m-%d')
            num_day = (lst_date - fst_date).days
            days_in_bin = int(num_day / self.n_bin)
            date_list = [fst_date]
            tmp_date = fst_date
            for i in range(self.n_bin):
                tmp_date = tmp_date + timedelta(days=days_in_bin)
                date_list.append(tmp_date)
            date_list[-1] = lst_date
            date_list = [x.strftime('%Y-%m-%d') for x in date_list]
            self.time_range_list = date_list
        print('时间窗口如下:')
        print(self.time_range_list)

    def run(self):
        for ft in self.ft_list:
            res_list = []
            n_sample_list = []
            miss_rate_list = []
            for i in range(len(self.time_range_list) - 1):
                tmp_df = self.df.loc[(self.df[self.time_field] >= self.time_range_list[i]) &
                                     (self.df[self.time_field] < self.time_range_list[i + 1])]

                tmp_df_notna = tmp_df.loc[tmp_df[ft].notna()]
                n_sample_list.append(tmp_df.shape[0])
                miss_rate_list.append(tmp_df[ft].isnull().mean())

                if self.eval_metric == 'auc':
                    try:
                        res_list.append(roc_auc_score(y_true=tmp_df_notna[self.label],
                                                      y_score=tmp_df_notna[ft]))
                    except:
                        res_list.append(0)
                else:
                    pass
            self.plot(res_list, ft, n_sample_list=n_sample_list, miss_rate_list=miss_rate_list)

    def plot(self, data, title,
             n_sample_list=None, miss_rate_list=None):
        x_tick = [-0.5]
        for i in range(1, len(data) + 1):
            x_tick.append(i - 0.5)
        plt.figure(figsize=(12, 8))
        plt.style.use('ggplot')

        stats_info = ''
        for i in range(len(self.time_range_list) - 1):
            stats_info += '%s~%s 样本量%d 缺失率%.4f %s%.3f\n' % (self.time_range_list[i],
                                                            self.time_range_list[i + 1],
                                                            n_sample_list[i], miss_rate_list[i],
                                                            self.eval_metric, data[i])
        plt.plot(list(range(len(data))), data, label=stats_info)
        plt.legend(fontsize=15)
        plt.title(title, fontsize=25)
        plt.xticks(x_tick, self.time_range_list, fontsize=15, rotation=-15)
        plt.yticks(fontsize=15)
        plt.subplots_adjust(left=0.1, bottom=0.1, right=0.9, top=0.9)
        plt.savefig(self.save_path + title)
        plt.show()


class EvaluationOnGroup:
    def __init__(self, df=None, label='target',
                 group_field=None, group_list=None,
                 eval_metric='auc', save_path='./image/EvaluationOnGroup/',
                 map_dict=None):
        self.df = df
        self.label = label
        self.group_field = group_field
        self.group_list = group_list
        self.eval_metric = eval_metric
        self.save_path = save_path
        mkdir(self.save_path)

        self.ft_list = self.df.columns.tolist()
        if self.label in self.ft_list:
            self.ft_list.remove(self.label)
        else:
            print('警告, 数据不包含目标列.')
        if self.group_field in self.ft_list:
            self.ft_list.remove(self.group_field)
        else:
            print('警告, 数据不包含分组列.')
        if self.group_list is None:
            self.group_list = self.df[self.group_field].unique().tolist()

        self.map_dict = map_dict
        if self.map_dict is not None:
            self.group_list_map = []
            for g in self.group_list:
                self.group_list_map.append(self.map_dict[g])
        else:
            self.group_list_map = self.group_list

    def run(self):
        for ft in self.ft_list:
            res_list = []
            n_sample_list = []
            miss_rate_list = []
            for group in self.group_list:
                tmp_df = self.df.loc[self.df[self.group_field] == group]
                tmp_df_notna = tmp_df.loc[tmp_df[ft].notna()]

                if self.eval_metric == 'auc':
                    try:
                        res_list.append(roc_auc_score(y_true=tmp_df_notna[self.label],
                                                      y_score=tmp_df_notna[ft]))
                    except:
                        res_list.append(0)
                else:
                    pass
                n_sample_list.append(tmp_df.shape[0])
                miss_rate_list.append(tmp_df[ft].isnull().mean())
            self.plot(res_list, ft, n_sample_list=n_sample_list, miss_rate_list=miss_rate_list)

    def plot(self, data, title,
             n_sample_list=None, miss_rate_list=None):
        plt.figure(figsize=(12, 8))
        plt.style.use('ggplot')
        stats_info = ''
        for i in range(len(self.group_list)):
            stats_info += '%s 样本量%d 缺失率%.4f %s%.3f\n' % (self.group_list[i],
                                                         n_sample_list[i], miss_rate_list[i],
                                                         self.eval_metric, data[i])
        plt.plot(list(range(len(data))), data, label=stats_info)
        plt.legend(fontsize=15)
        plt.title(title, fontsize=25)
        plt.xticks(list(range(len(data))), self.group_list_map, fontsize=15, rotation=-15)
        plt.yticks(fontsize=15)
        plt.subplots_adjust(left=0.1, bottom=0.1, right=0.9, top=0.9)
        plt.savefig(self.save_path + title)
        plt.show()


class Mchart:
    def __init__(self,
                 save_path='./image/'):
        self.save_path = save_path
        mkdir(save_path)

    def hist(self, x=None, bins=20,
             hist=True, kde=True, figsize=(12, 8),
             title='test', x_name=None,
             is_x_time=False, time_fmt='%Y-%m-%d'):
        plt.figure(figsize=figsize)
        if is_x_time:
            # 把日期转换为时间戳.
            try:
                x = [datetime.strptime(i, time_fmt) for i in x]
            except:
                pass
            x = [i.timestamp() for i in x]
            # 利用时间戳画直方图.
            sns.distplot(a=x, bins=bins, hist=hist, kde=kde)

            # 修改直方图横轴刻度.
            x_ticks = [np.percentile(x, i) for i in [0, 20, 40, 60, 80, 100]]
            x_label = [datetime.fromtimestamp(i).strftime(time_fmt) for i in x_ticks]
            plt.xticks(x_ticks, x_label, rotation=-15)
        else:
            sns.distplot(a=x, bins=bins, hist=hist, kde=kde)
            plt.xlabel(xlabel=x_name, fontsize=15)
        # 配置.
        plt.title(title, fontsize=25)
        plt.xticks(fontsize=15)
        plt.yticks(fontsize=15)
        plt.subplots_adjust(left=0.1, bottom=0.1, right=0.9, top=0.9)
        plt.savefig(self.save_path + title)
        plt.show()


class Echart:
    def __init__(self,
                 mode='html',
                 save_path='./image/'):
        '''
        :param mode: str, 指定模式,
            默认html, 可选notebook(nb), static.
        '''

        self.mode = mode
        self.save_path = save_path

        mkdir(save_path)

    def output(self, painter, title='test'):
        if self.mode == 'html':
            painter.render(path=self.save_path + title + '.html')
        elif self.mode == 'nb':
            painter.render_notebook()
        elif self.mode == 'static':
            make_snapshot(engine=snapshot,
                          file_name=painter.render(),
                          output_name=self.save_path + title + '.png',
                          pixel_ratio=2,
                          is_remove_html=True)
        else:
            print('模式输入错误, 可选模式: \nhtml\nnb\nstatic')

    def line(self, x=None, y=None, title='test',
             is_max_show=False, is_min_show=False,
             is_avg_show=False, is_x_time=False,
             x_name='X', y_name='Y',
             y_name_list=None):
        # 初始化.
        painter = Line(init_opts=opts.InitOpts(width='900px',
                                               height='500px',
                                               page_title=title))
        # 配置.
        painter.set_global_opts(title_opts=opts.TitleOpts(title=title),
                                tooltip_opts=opts.TooltipOpts(trigger='axis',
                                                              axis_pointer_type='line'),
                                xaxis_opts=opts.AxisOpts(
                                    type_='time' if is_x_time else ('category' if isinstance(x[0], str) else 'value'),
                                    name=x_name,
                                    is_scale=True),
                                yaxis_opts=opts.AxisOpts(name=y_name,
                                                         is_scale=True),
                                legend_opts=opts.LegendOpts(is_show=True if isinstance(y[0], list) else False))
        # 添加数据.
        painter.add_xaxis(xaxis_data=x)
        if isinstance(y[0], list):
            for i, y_ in enumerate(y):
                painter.add_yaxis(series_name=y_name_list[i] if y_name_list is not None else '',
                                  y_axis=y_,
                                  is_connect_nones=True,
                                  label_opts=opts.LabelOpts(is_show=False),
                                  symbol_size=10)
        else:
            painter.add_yaxis(series_name='',
                              y_axis=y,
                              is_connect_nones=True,
                              label_opts=opts.LabelOpts(is_show=False),
                              symbol_size=10,
                              markline_opts=opts.MarkLineOpts(
                                  data=[opts.MarkAreaItem(type_='average')]) if is_avg_show else None,
                              markpoint_opts=opts.MarkPointOpts(
                                  data=[opts.MarkPointItem(type_='max' if is_max_show else 'min')]) if (
                                      is_max_show or is_min_show) else None)

        # 输出.
        self.output(painter=painter, title=title)

    def bar(self, x=None, y=None, title='test',
            is_max_show=False, is_min_show=False,
            is_reversal=False, is_stack=False,
            is_order=False,
            x_name='X', y_name='Y',
            y_name_list=None):
        # 数据预处理.
        if is_order:
            x_y = sorted(zip(x, y), key=lambda p: p[1], reverse=True)
            x = [x[0] for x in x_y]
            y = [x[1] for x in x_y]

        # 初始化.
        painter = Bar(init_opts=opts.InitOpts(width='900px',
                                              height='500px',
                                              page_title=title))

        # 配置.
        painter.set_global_opts(title_opts=opts.TitleOpts(title=title),
                                xaxis_opts=opts.AxisOpts(name=x_name,
                                                         is_scale=True),
                                yaxis_opts=opts.AxisOpts(name=y_name,
                                                         is_scale=True),
                                legend_opts=opts.LegendOpts(is_show=True if isinstance(y[0], list) else False))

        # 添加数据.
        painter.add_xaxis(xaxis_data=x)
        if isinstance(y[0], list):
            for i, y_ in enumerate(y):
                painter.add_yaxis(series_name=y_name_list[i] if y_name_list is not None else '',
                                  yaxis_data=y_,
                                  gap='0%' if not is_stack else None,
                                  stack='stack1' if is_stack else None)

        else:
            painter.add_yaxis(series_name='', yaxis_data=y,
                              markpoint_opts=opts.MarkPointOpts(
                                  data=[opts.MarkPointItem(name='MAX' if is_max_show else 'MIN',
                                                           type_='max' if is_max_show else 'min')]) if (
                                      is_max_show or is_min_show) else None)
        if is_reversal:
            painter.reversal_axis()
        if is_stack:
            painter.set_series_opts(label_opts=opts.LabelOpts(is_show=False))
        if is_reversal and not is_stack:
            painter.set_series_opts(label_opts=opts.LabelOpts(position='right'))

        # 输出.
        self.output(painter=painter, title=title)

    def hist(self, x=None, title='test', bins=20,
             x_name='X'):
        # 预处理.
        df = pd.DataFrame({'interval': pd.cut(x, bins=bins).tolist()})
        df['mean'] = df['interval'].apply(lambda x: (x.left + x.right) / 2)
        df_g = df.groupby('mean').agg(count_=('interval', 'count'))
        x = [round(i, 4) for i in df_g.index.tolist()]
        y = df_g['count_'].tolist()

        # 初始化.
        painter = Bar(init_opts=opts.InitOpts(width='900px',
                                              height='500px',
                                              page_title=title))

        # 配置.
        painter.set_global_opts(title_opts=opts.TitleOpts(title=title),
                                xaxis_opts=opts.AxisOpts(is_scale=True, name=x_name),
                                yaxis_opts=opts.AxisOpts(is_scale=True),
                                legend_opts=opts.LegendOpts(is_show=False))

        # 添加数据.
        painter.add_xaxis(xaxis_data=x)
        painter.add_yaxis(series_name='', yaxis_data=y,
                          label_opts=opts.LabelOpts(is_show=False),
                          category_gap=0)

        # 输出.
        self.output(painter=painter, title=title)

        pass


def EDA(df=None, title='EDA.html', max_sample=10000):
    # 若数据过大, 则抽样后再进行报告生成.
    if df.shape[0] > max_sample:
        df = df.sample(n=max_sample)
    profile = df.profile_report(title=title)
    mkdir('./log/')
    profile.to_file(output_file='./log/' + title)
    print('./log/' + title + '生成完毕.')


def make_html_report(image_dir_path='./image/',
                     md_html_path='md.html'):
    '''
    把echart图片插进markdown生成的html文件.
    '''
    print('开始插入html.')
    # 加载js语句.
    javascript = '''<script type="text/javascript" src="https://assets.pyecharts.org/assets/echarts.min.js"></script>'''
    # 读取md_html文本.
    with open(md_html_path, 'r', encoding='utf-8') as f:
        html = f.read()
        f.close()
    # 添加js语句.
    html = html.replace(r'''<meta charset='UTF-8'>''',
                        r'''<meta charset='UTF-8'><meta name="viewport" content="width=device-width, initial-scale=1">''' + javascript)
    # 遍历指定图片路径中的图片, 将图片插入html.
    for file in os.listdir(image_dir_path):
        if file.endswith('.html'):
            if '<p>@' + file[: -5] + '@</p>' in html:
                with open(image_dir_path + file, 'r', encoding='utf-8') as f:
                    content = f.read()
                    f.close()
                    # 匹配.
                    image_js = re.search(r'<div.*</script>', content, flags=re.M | re.DOTALL).group()

                    # 替换.
                    html = html.replace('<p>@' + file[: -5] + '@</p>', image_js)

    # 保存.
    with open(md_html_path, 'w', encoding='utf-8') as f:
        f.write(html)
        f.close()
    print('插入完毕.')


if __name__ == '__main__':
    make_html_report(md_html_path='首贷无运营商-融合模型-说明.html')

    pass



# coding=utf-8
'''
流程整合.
减少灵活性, 增加便利性.
'''
import gc
import pandas as pd
from sklearn.metrics import roc_auc_score

from .data_transform import LinearModelPreprocessor
from .param_optimize import LightBayesOptimizer, LROptimizer
from .feature_select import FeatureSelector
from .base_utils import mkdir, split_X_y
from .feature_select import ModelParamsGenerator, LGBRFE

def gen_model_params_and_ft_select(df=None, label='target'):
    mpg = ModelParamsGenerator(df=df, label=label)
    model_params_dict = mpg.run()

    lgbrfe = LGBRFE(df=df, label=label, params=model_params_dict)
    lgbrfe.run()
    lgbrfe.plot()



def ft_select_and_param_opt(select_method='null_imp', train=None,
                            test=None, label='target', time_filed=None,
                            is_pred=False, name='origin', n_iter=50):
    '''
    将特征选择与超参数调节, 模型评估放在一起.
    若没有测试集, 则保存训练好的模型.
    若没有测试集标签, 则进一步保存并输出测试集预测结果.
    若有测试集标签, 则输出测试集AUC.
    :param select_method: str, 特征选择的方法, 默认"null_imp".
    :param train: df, 训练集.
    :param test: df, 测试集.
    :param label: str, 标签列, 默认"target".
    :param time_filed: str, 时间列.
    :param is_pred: bool, 是否输出测试集预测结果, 默认False, 即输出AUC.
    :param name: str, 名称标准, 以区分不同的数据.
    :param n_iter: int, 超参数随机搜索次数, 默认20.
    :return: null or float, 根据模式不返回或返回AUC.
    '''
    # 创建文件夹保存信息.
    mkdir('./feature/')
    mkdir('./model/')
    mkdir('./result/')
    ###############################################################
    # 参数选择.
    if select_method == 'null_imp':
        if time_filed is None:
            fs = FeatureSelector(df=train, label=label)
        else:
            fs = FeatureSelector(df=train.drop(time_filed, axis=1), label=label)
        fs.null_imp_select(verbose=0)
        fs.save_feature(path='./feature/' + name + '_' + select_method,
                        is_save_ft_list=False, is_save_ft_list_dict=True)
        ft_list_dict = fs.select_ft_list_dict.copy()
    elif select_method == 'boruta':
        if time_filed is None:
            fs = FeatureSelector(df=train, label=label)
        else:
            fs = FeatureSelector(df=train.drop(time_filed, axis=1), label=label)
        fs.boruta_select(verbose=0)
        fs.save_feature(path='./feature/' + name + '_' + select_method,
                        is_save_ft_list=False, is_save_ft_list_dict=True)
        ft_list_dict = fs.select_ft_list_dict.copy()
    elif select_method == 'eli5':
        if time_filed is None:
            fs = FeatureSelector(df=train, label=label)
        else:
            fs = FeatureSelector(df=train.drop(time_filed, axis=1), label=label)
        fs.eli5_permutation_select()
        fs.save_feature(path='./feature/' + name + '_' + select_method,
                        is_save_ft_list=False, is_save_ft_list_dict=True)
        ft_list_dict = fs.select_ft_list_dict.copy()
    elif select_method == 'gain':
        if time_filed is None:
            fs = FeatureSelector(df=train, label=label)
        else:
            fs = FeatureSelector(df=train.drop(time_filed, axis=1), label=label)
        fs.gain_imp_select()
        fs.save_feature(path='./feature/' + name + '_' + select_method,
                        is_save_ft_list=False, is_save_ft_list_dict=True)
        ft_list_dict = fs.select_ft_list_dict.copy()
    # 释放内存.
    del fs
    gc.collect()

    ##############################################################
    # 模型训练.

    for p in ft_list_dict:
        ft_list = ft_list_dict[p]
    ##############################################################
        # cv logloss.
        lgb_opt = LightBayesOptimizer(df=train[ft_list + [label]], time_filed=None,
                                      label=label, opt_mode='cv', metrics='logloss')
        lgb_opt.run(init_points=n_iter, n_iter=0, verbose=0, is_verbose=0)
        lgb_opt.fit_model()
        # 保存模型.
        lgb_opt.save_model(path='./model/' + name + '_' + select_method + '_'
                                + str(p) + '_' + 'cv_logloss')
        # 模型预测/评估.
        if test is None:
            pass
        else:
            if is_pred:
                result = lgb_opt.predict(test)
                result = pd.DataFrame({'predict': result})
                result.to_csv('./result/' + name + '_' + select_method + '_'
                              + str(p) + '_' + 'cv_logloss.csv')
            else:
                print(name + '_' + select_method + '_'
                      + str(p) + '_' + 'cv_logloss AUC is : %f' %
                      roc_auc_score(test[label], lgb_opt.predict(test)))
    ##############################################################
        # ho logloss
        lgb_opt = LightBayesOptimizer(df=train[ft_list + [label]], time_filed=None,
                                      label=label, opt_mode='ho', metrics='logloss')
        lgb_opt.run(init_points=n_iter, n_iter=0, verbose=0, is_verbose=0)
        lgb_opt.fit_model()
        # 保存模型.
        lgb_opt.save_model(path='./model/' + name + '_' + select_method + '_'
                                + str(p) + '_' + 'ho_logloss')
        # 模型预测/评估.
        if test is None:
            pass
        else:
            if is_pred:
                result = lgb_opt.predict(test)
                result = pd.DataFrame({'predict': result})
                result.to_csv('./result/' + name + '_' + select_method + '_'
                              + str(p) + '_' + 'ho_logloss.csv')
            else:
                print(name + '_' + select_method + '_'
                      + str(p) + '_' + 'ho_logloss AUC is : %f' %
                      roc_auc_score(test[label], lgb_opt.predict(test)))
    ##############################################################
        # ho logloss time
        if time_filed is not None:
            lgb_opt = LightBayesOptimizer(df=train[ft_list + [label, time_filed]], time_filed=time_filed,
                                          label=label, opt_mode='ho', metrics='logloss')
            lgb_opt.run(init_points=n_iter, n_iter=0, verbose=0, is_verbose=0)
            lgb_opt.fit_model()
            # 保存模型.
            lgb_opt.save_model(path='./model/' + name + '_' + select_method + '_'
                                    + str(p) + '_' + 'ho_logloss_time')
            # 模型预测/评估.
            if test is None:
                pass
            else:
                if is_pred:
                    result = lgb_opt.predict(test)
                    result = pd.DataFrame({'predict': result})
                    result.to_csv('./result/' + name + '_' + select_method + '_'
                                  + str(p) + '_' + 'cv_logloss_time.csv')
                else:
                    print(name + '_' + select_method + '_'
                          + str(p) + '_' + 'cv_logloss_time AUC is : %f' %
                          roc_auc_score(test[label], lgb_opt.predict(test)))


def linear_model_preprocess_and_fit(train=None,
                                    test=None,
                                    label='target',
                                    is_boxcox=True,
                                    boxcox_mode='yeo-johnson',
                                    is_miss_impute=True,
                                    min_miss_num=100,
                                    miss_mode='mean',
                                    is_indicate=True,
                                    nfold=7,
                                    Cs=10):
    '''
    :param train: 数据集.
    :param label: 标签名称.
    :param is_boxcox: 是否BoxCox.
    :param boxcox_mode: boxcox模式.
    :param is_miss_impute: 是否填充缺失值.
    :param min_miss_num: 最小缺失填充阈值.
    :param miss_mode: 填充缺失值模式.
    :param is_indicate: 是否添加指示列.
    :param nfold: 交叉验证折数.
    :param Cs: 正则化系数.
    :return: 训练好的逻辑回归模型.
    '''
    # 数据预处理.
    lmp = LinearModelPreprocessor(is_boxcox=is_boxcox,
                                  boxcox_mode=boxcox_mode,
                                  is_miss_impute=is_miss_impute,
                                  min_miss_num=min_miss_num,
                                  miss_mode=miss_mode,
                                  is_indicate=is_indicate)
    train_X, train_y = split_X_y(train, label=label)
    test_X, test_y = split_X_y(test, label=label)

    train_X = lmp.fit_transform(X=train_X, y=train_y)
    test_X = lmp.transform(X=test_X)
    train_X_y, test_X_y = train_X, test_X
    train_X_y[label], test_X_y[label] = train_y, test_y
    del train_X, test_X
    gc.collect()

    # 逻辑回归调参.
    lro = LROptimizer(df=train_X_y, label=label, nfold=nfold, Cs=Cs)
    lro.run()

    # 输出结果.
    print('训练集结果:')
    print(roc_auc_score(train_X_y[label], lro.predict(train_X_y)))
    print('测试集结果.')
    print(roc_auc_score(test_X_y[label], lro.predict(test_X_y)))

    return lmp, lro


if __name__ == '__main__':

    train = pd.read_csv('./data/bc_train.csv')
    gen_model_params_and_ft_select(df=train)



# coding=utf-8
'''
工具.
'''
import os
import pickle
import warnings
import time
import numpy as np
import pandas as pd
from collections import defaultdict
from itertools import combinations
from sklearn.utils.class_weight import compute_sample_weight
from sklearn.metrics import log_loss
from hashlib import md5


warnings.filterwarnings('ignore')


def save(path='smileshy.pkl', obj=None):
    prefix = '/'.join(path.split('/')[: -1])
    if len(prefix) == 0:
        pass
    else:
        mkdir(prefix)
    with open(path, 'wb') as f:
        pickle.dump(obj, f)
        f.close()
    print('保存%s完毕!' % path)


def load(path='smileshy.pkl'):
    with open(path, 'rb') as f:
        obj = pickle.load(f)
        f.close()
    print('加载%s完毕!' % path)
    return obj


def mkdir(path='./path'):
    '''
    根据给定的路径创建目录.
    '''
    os.makedirs(path.encode('utf-8'), exist_ok=True)


def timer(func):
    '''
    计算时间的装饰器.
    '''

    def wraper(*args, **kwargs):
        start_time = time.time()
        print('=' * 40)
        print('%s开始执行' % func.__doc__)
        result = func(*args, **kwargs)
        end_time = time.time()
        time_long = end_time - start_time
        if time_long < 3600:
            print('%s执行完毕, 花费时间为: %f 分钟.' % (func.__name__, round(time_long / 60, 2)))
        else:
            print('%s执行完毕, 花费时间为: %f 小时.' % (func.__name__, round(time_long / 3600, 2)))
        print('=' * 40)
        return result
    return wraper


def random_num_generator(n=1000, random_state=777):
    '''
    随机数生成器, 范围0-1.
    '''
    np.random.seed(random_state)
    for i in range(n):
        yield np.random.rand()


def make_miss_value(df=None, label='target', row_perc=0.2, col_perc=0.8, random_state=7):
    '''
    随机将原数据中部分数据设置为NaN.
    '''
    ft_list = df.columns.tolist()
    if label in ft_list:
        ft_list.remove(label)
    np.random.seed(7)
    np.random.shuffle(ft_list)
    ft_miss_list = []
    for i in range(int(len(ft_list) * col_perc)):
        ft_miss_list.append(ft_list.pop())
    for i, ft in enumerate(ft_miss_list):
        rand_num = random_num_generator(n=df.shape[0], random_state=i)
        df[ft] = [np.nan if rand_num.__next__() < row_perc else x for x in df[ft]]


def split_X_y(df=None, label='target'):
    if label in df.columns.tolist():
        return df.drop(label, axis=1).reset_index(drop=True), df[label].values
    else:
        print('数据集中没有目标列.')


def rank_sort(data):
    len_data = len(data)
    sort_data = sorted(data)
    rank_data = list(range(1, len_data + 1))
    map_dict = defaultdict(float)
    count_dict = defaultdict(int)
    for i, d in enumerate(sort_data):
        map_dict[d] += rank_data[i]
        count_dict[d] += 1
    for k in count_dict:
        map_dict[k] /= count_dict[k]
    return [map_dict[x] for x in data]


def rank_sort_combine(data_list, weight=None):
    rank_data_list = []
    for data in data_list:
        rank_data_list.append(rank_sort(data))
    res = 0
    if weight is None:
        for data in rank_data_list:
            res = res + np.array(data)
        return res
    else:
        for i, data in enumerate(rank_data_list):
            res = res + np.array(data) * weight[i]
        return res


def traversal_search(data_list, n_range=(1, 2)):
    for n in n_range:
        for cb in combinations(data_list, n):
            yield list(cb)


def neg_log_loss(y_true, y_pred):
    weight = compute_sample_weight(class_weight='balanced', y=y_true)
    return -log_loss(y_true=y_true, y_pred=y_pred, sample_weight=weight)


def clean_float_data(data=None, data_type='df'):
    def clean_data(data):
        try:
            return float(data)
        except:
            return np.nan

    if data_type == 'df':
        for ft in data.columns.tolist():
            data[ft] = data[ft].apply(clean_data)
    elif data_type == 'list':
        data = map(clean_data, data)
    return data


def md5_encode(data):
    return md5(data.encode()).hexdigest()


def equal_frequency_bin(data, n_bins=10):
    '''
    等频分箱.
    返回分箱后的编码和边界.
    '''
    return pd.qcut(data, n_bins, labels=False, retbins=True, precision=8, duplicates='drop')



if __name__ == '__main__':
    a=['a','b','c', 'w']
    for i in traversal_search(a, (1, 2, 3, 4)):
        print(i)
    pass


# coding=utf-8
'''
暴力特征衍生.
根据给定的特征对象在原始数据中计算特征.
'''

import pandas as pd
import numpy as np
from datetime import datetime, timedelta
import featuretools

import warnings
warnings.filterwarnings('ignore')
from .base_utils import mkdir, save, timer, load

class SimpleFeatureGenerator:
    def __init__(self, gap_fill=''):
        self.gap_fill = gap_fill
        self.ft_obj_dict = {}
        self.trans_map_dict = {'*': 'multiply_numeric',
                               '/': 'divide_numeric',
                               '+': 'add_numeric',
                               '-': 'subtract_numeric'}
        self.agg_map_dict = {'max': 'max',
                             'min': 'min'}

    @timer
    def numerical_generate(self, df=None, label='target',
                           trans_types=('/', '*'),
                           agg_type=('min', 'max')):
        '''
        对数值型特征进行衍生.
        '''
        if label in df.columns.tolist():
            del df[label]
        trans_type_list = [self.trans_map_dict[x] if x in self.trans_map_dict else x for x in trans_types]
        agg_type_list = [self.agg_map_dict[x] if x in self.agg_map_dict else x for x in agg_type]
        es = featuretools.EntitySet(id='full_data')
        es.entity_from_dataframe(entity_id='data', dataframe=df,
                                 index='idx', make_index=True)
        ft_df, ft_name = featuretools.dfs(entityset=es, target_entity='data',
                                          trans_primitives=trans_type_list,
                                          agg_primitives=agg_type_list,
                                          max_depth=1,
                                          n_jobs=4)
        ft_df.columns = [self.gap_fill.join(x.split(' ')) for x in ft_df.columns]
        self.ft_obj_dict = dict(zip(ft_df.columns.tolist(), ft_name))

        return ft_df

    @timer
    def feature_generate(self, df=None, ft_list=None, label='target'):
        '''
        给定衍生后的特征, 从原始数据中进行计算.
        '''
        # 确定要原始数据集, 以及要计算的特征.
        if label in df.columns.tolist():
            del df[label]
        ft_obj_list = []
        for ft in ft_list:
            ft_obj_list.append(self.ft_obj_dict[ft])
        # 计算特征.
        es = featuretools.EntitySet(id='full_data')
        es.entity_from_dataframe(entity_id='data', dataframe=df,
                                 index='idx', make_index=True)
        ft_df = featuretools.calculate_feature_matrix(features=ft_obj_list,
                                                      entityset=es,
                                                      n_jobs=4)
        return ft_df

    def save(self, save_path='./model/ft_generator.obj'):
        save(save_path, self)
        print(save_path + ' 保存成功.')


class TimeFeatureGenerator:
    def __init__(self):

        pass

    def hour_generate(self, df=None, time_field='time',
                      time_interval=1,
                      time_fmt='%Y-%m-%d %H:%M:%S'):
        # 转换时间格式.
        time_data = df[time_field].copy()
        try:
            time_data = time_data.apply(lambda x: datetime.strptime(x, time_fmt))
        except:
            pass
        # 制作编码字典.
        time_delta = timedelta(hours=time_interval)
        start_time = datetime.strptime('00:00:00', '%H:%M:%S')
        end_time = datetime.strptime('23:59:59', '%H:%M:%S')
        time_interval_list = []
        for i in range(int(24 / time_interval) - 1):
            time_interval_list.append((start_time, start_time + time_delta))
            start_time = start_time + time_delta
        time_interval_list.append((start_time, end_time))
        # 按小时编码.
        time_encode = []
        for t in time_data:
            if pd.isnull(t):
                time_encode.append(np.nan)
                continue
            for i, t_interval in enumerate(time_interval_list):
                if t.time() >= t_interval[0].time() and t.time() < t_interval[1].time():
                    time_encode.append(str(i))
                    break
        df['hour_encode_' + str(time_interval)] = time_encode
        print(df['hour_encode_' + str(time_interval)].value_counts())
        return df

    def week_genetare(self):

        pass







if __name__ == '__main__':
    # from sklearn.datasets import load_breast_cancer
    #
    # data, target = load_breast_cancer(return_X_y=True)
    # df = pd.DataFrame(data, columns=[str(x) for x in range(data.shape[1])])
    # df['target'] = target
    # train, test = train_test_split(df, random_state=777, stratify=df['TARGET'], test_size=0.2)
    # train = train.reset_index(drop=True)
    # test = test.reset_index(drop=True)

    df = pd.DataFrame({'a':[1,2,3],
                       'b':[2,3,4],
                       'c':[4,5,6]})
    fg = SimpleFeatureGenerator()
    c = df['c'].tolist()
    df = fg.numerical_generate(df=df, label='c')
    print(df)
    df['c'] = c
    print(df)



# coding=utf-8
import setuptools

setuptools.setup(name='shy-tools',
                 version='0.0.9',
                 author='shy',
                 author_email='13141462249@163.com',
                 url='https://github.com/smileshy777/shy_tools',
                 description='My code tools',
                 packages=setuptools.find_packages())
